/* Written by Yu-Fang Chen, Richard Mayr, and Chih-Duo Hong               */
/* Copyright (c) 2010                  	                                  */
/*                                                                        */
/* This program is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation; either version 2 of the License, or      */
/* (at your option) any later version.                                    */
/*                                                                        */
/* This program is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */
/*                                                                        */
/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA*/

package algorithms;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import automata.FAState;
import automata.FiniteAutomaton;
import comparator.GraphComparator;
import comparator.MetaGraphComparator;
import comparator.StatePairComparator;
import comparator.SuperGraphComparator;
import datastructure.Arc;
import datastructure.HashSet;
import datastructure.OneToOneTreeMap;
import datastructure.Pair;



/**
 * 
 * @author Yu-Fang Chen
 * 
 */
public class InclusionOpt3 extends Thread{
	public boolean backward=true;
	public boolean C1=true;
	public boolean C2=true;
	public boolean opt2=true;
	public boolean quotient=false;
	public boolean debug=true;
	public int removedCnt;
	public boolean included=true;

	
	
	TreeMap<TreeSet<Arc>,TreeSet<Arc>> R=new TreeMap<TreeSet<Arc>,TreeSet<Arc>>(new GraphComparator());
	TreeMap<TreeSet<Arc>,TreeSet<Arc>> L=new TreeMap<TreeSet<Arc>,TreeSet<Arc>>(new GraphComparator());
	
	private TreeMap<Integer,TreeSet<FAState>> Post=new TreeMap<Integer,TreeSet<FAState>>();
	private TreeMap<Integer,TreeSet<FAState>> NPost=new TreeMap<Integer,TreeSet<FAState>>();

	private TreeMap<String,TreeSet<Integer>> Tail=new TreeMap<String,TreeSet<Integer>>();
	private TreeMap<String,TreeSet<Integer>> Head=new TreeMap<String,TreeSet<Integer>>();
	private long runTime;
	private boolean stop=false;
	private Set<Pair<FAState,FAState>> frel, brel;
	FiniteAutomaton spec, system;
	String showSim(Set<Pair<FAState,FAState>> sim){
		String result="";
		Iterator<Pair<FAState,FAState>> sim_it=sim.iterator();
		while(sim_it.hasNext()){
			Pair<FAState,FAState> p=sim_it.next();
			if(p.getLeft().compareTo(p.getRight())!=0)
				result+=("("+p.getLeft()+","+p.getRight()+")");
		}
		return result;
	}
	void debug(String out){
		if(debug)
		System.out.println(out);
	}
	
	
	public InclusionOpt3(FiniteAutomaton system, FiniteAutomaton spec){
		this.spec=spec;
		this.system=system;
	}
	/**
	 * Simplify a finite automaton by merging simulation equivalent states
	 * @param fa: a finite automaton
	 * @param Sim: some simulation rel_specation over states in the spec automaton
	 * 
	 * @return an equivalent finite automaton
	 */
	private FiniteAutomaton quotient(FiniteAutomaton fa, Set<Pair<FAState,FAState>> rel) {
		FiniteAutomaton result=new FiniteAutomaton();
		result.name=fa.name;
		TreeMap<FAState,FAState> map=new TreeMap<FAState,FAState>();
		TreeMap<FAState,FAState> reducedMap=new TreeMap<FAState,FAState>();
		
		Iterator<FAState> state_it=fa.states.iterator();
		while(state_it.hasNext()){
			FAState state=state_it.next();
			map.put(state, state);
			Iterator<FAState> state_it2=fa.states.iterator();
			while(state_it2.hasNext()){
				FAState state2=state_it2.next();
				if(rel.contains(new Pair<FAState,FAState>(state,state2)) &&
					rel.contains(new Pair<FAState,FAState>(state2,state))){
					map.put(state,state2);
				}
			}			
		}

		FAState init=result.createState();
		reducedMap.put(map.get(fa.getInitialState()), init);
		result.setInitialState(init);

		state_it=fa.states.iterator();
		while(state_it.hasNext()){
			FAState state=state_it.next();
			if(!reducedMap.containsKey(map.get(state))){
				reducedMap.put(map.get(state), result.createState());
			}
			if(fa.F.contains(state)){
				result.F.add(reducedMap.get(map.get(state)));
			}
			Iterator<String> sym_it=state.nextIt();
			while(sym_it.hasNext()){
				String sym=sym_it.next();
				Iterator<FAState> to_it=state.getNext(sym).iterator();
				while(to_it.hasNext()){
					FAState to=to_it.next();
					if(!reducedMap.containsKey(map.get(to))){
						reducedMap.put(map.get(to), result.createState());
					}
					result.addTransition(reducedMap.get(map.get(state)), reducedMap.get(map.get(to)), sym);
				}
			}
		}
		Set<Pair<FAState,FAState>> newrel=new TreeSet<Pair<FAState,FAState>>(new StatePairComparator());
		Iterator<Pair<FAState,FAState>> sim_it=rel.iterator();
		while(sim_it.hasNext()){
			Pair<FAState,FAState> sim=sim_it.next();
			FAState left,right;
			if(sim.getLeft().getowner()==fa){
				left=reducedMap.get(map.get(sim.getLeft()));
			}else{
				left=sim.getLeft();
			}
			
			if(sim.getRight().getowner()==fa){
				right=reducedMap.get(map.get(sim.getRight()));
			}else{
				right=sim.getRight();
			}
			newrel.add(new Pair<FAState,FAState>(left,right));
		}
		rel.clear();
		rel.addAll(newrel);
		
		return result;
	}	

	private boolean C1(Arc left,TreeSet<Arc> right){
		if(!C1){
			return false;
		}
		if(left.getFrom()!=system.getInitialState().id){
			debug("C1("+left+","+right+")="+true);
			return true;
		}
		Iterator<Arc> right_it=right.iterator();
		while(right_it.hasNext()){
			Arc r=right_it.next();
			if(r.getFrom()==spec.getInitialState().id){
				if(frel.contains(new Pair<FAState,FAState>(new FAState(left.getTo(),system),new FAState(r.getTo(),spec)))){
					debug("C1("+left+","+right+")="+true);
					return true;
				}
			}
		}
		debug("C1("+left+","+right+")="+false);
		return false;
	}

	private TreeSet<FAState> getPost(int qid){
		TreeSet<FAState> post=Post.get(qid);
		if(post==null){
			post=new TreeSet<FAState>();
			FAState q=null;
			Iterator<FAState> st_it=system.states.iterator();
			while(st_it.hasNext()){
				FAState st=st_it.next();
				if(st.getID()==qid){
					q=st;
					break;
				}
			}
			Set<FAState> next=q.getNext();
			post.addAll(next);
			post.add(q);
			while(!next.isEmpty()){
				FAState s=next.iterator().next();
				next.remove(s);
				Iterator<FAState> toAdd_it=s.getNext().iterator();
				while(toAdd_it.hasNext()){
					FAState toAdd=toAdd_it.next();
					if(!post.contains(toAdd)){
						post.add(toAdd);
						next.add(toAdd);
					}
				}
			}
			Post.put(qid, post);
		}
		return post;
	}
	private boolean C2(Arc left, TreeSet<Arc> right){
		if(!C2){
			return false;
		}
		if(left.getLabel()){	
			TreeSet<FAState> post=getPost(left.getTo());
			Iterator<FAState> postst_it=post.iterator();
			while(postst_it.hasNext()){
				FAState postst=postst_it.next();
				if(frel.contains(new Pair<FAState,FAState>(new FAState(left.getFrom(),system),postst))){
					debug("C2("+left+","+right+")="+false);
					return false;
				}
			}
		}else{
			TreeSet<FAState> post=NPost.get(left.getTo());
			if(post==null){
				TreeSet<FAState> fpost=new TreeSet<FAState>();
				fpost.addAll(getPost(left.getTo()));
				fpost.retainAll(system.F);
				post=new TreeSet<FAState>();
				
				Iterator<FAState> st_it=fpost.iterator();
				while(st_it.hasNext()){
					FAState st=st_it.next();
					post.addAll(getPost(st.id));
				}
				NPost.put(left.getTo(), post);
			}
			Iterator<FAState> postst_it=post.iterator();
			while(postst_it.hasNext()){
				FAState postst=postst_it.next();
				if(frel.contains(new Pair<FAState,FAState>(postst,new FAState(left.getFrom(),system)))){
					debug("C2("+left+","+right+")="+false);
					return false;
				}
			}
		}
		debug("C2("+left+","+right+")="+true);
		return true;
	}
	
	private boolean double_graph_test(Pair<TreeSet<Arc>,TreeSet<Arc>> g, Pair<TreeSet<Arc>,TreeSet<Arc>> h, int init){
		Iterator<Arc> g_arcs_it=g.getLeft().iterator();
		while(g_arcs_it.hasNext()){
			Arc g_arc=g_arcs_it.next();
			if((C1||C2)&&(!L.containsKey(g.getRight())||!L.get(g.getRight()).contains(g_arc)))
				continue;
			Iterator<Arc> h_arcs_it=h.getLeft().iterator();	
			while(h_arcs_it.hasNext()){
				Arc h_arc=h_arcs_it.next();
				if((C1||C2)&&(!R.containsKey(h.getRight())||!R.get(h.getRight()).contains(h_arc)))
					continue;
				if(!(frel.contains(new Pair<FAState,FAState>(new FAState(h_arc.getFrom(),system),new FAState(h_arc.getTo(),system))) &&
						 frel.contains(new Pair<FAState,FAState>(new FAState(h_arc.getFrom(),system),new FAState(g_arc.getTo(),system))) &&
						 system.getInitialState().id==g_arc.getFrom() && h_arc.getLabel())
					)
					continue;
				boolean result;
				if(!backward)
					result = lasso_finding_test(g.getRight(),h.getRight(),init);
				else
					result = lasso_finding_test_plus(g.getRight(),h.getRight(),init);

				if(!result){
					debug("g_arc: "+g_arc);
					debug("h_arc: "+h_arc);
					debug("init: "+system.getInitialState());
					debug("final: "+system.F);
				}
				debug("RDGT("+g+"["+stringmap.get(g)+"],"+h+"["+stringmap.get(h)+"])="+result);
				return result;
			}
		}
		debug("RDGT("+g+"["+stringmap.get(g)+"],"+h+"["+stringmap.get(h)+"])="+true);
		return true;
	
	}	
		
	
	private boolean lasso_finding_test(TreeSet<Arc> g, TreeSet<Arc> h, int init){
		if(!Head.containsKey(g.toString())){
			TreeSet<Integer> H=new TreeSet<Integer>();
			Iterator<Arc> arc_g_it=g.iterator();
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				if(arc_g.getFrom()==init){
					H.add(arc_g.getTo());
				}
			}
			Head.put(g.toString(), H);
		}
		if(!Tail.containsKey(h.toString())){
			FiniteAutomaton fa=new FiniteAutomaton();
			OneToOneTreeMap<Integer,automata.FAState> st=new OneToOneTreeMap<Integer,automata.FAState>();
			Iterator<Arc> arc_h_it=h.iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				if(!st.containsKey(arc_h.getFrom()))
					st.put(arc_h.getFrom(), fa.createState());
				if(!st.containsKey(arc_h.getTo()))
					st.put(arc_h.getTo(), fa.createState());
				fa.addTransition(st.getValue(arc_h.getFrom()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");
			}
			SCC s=new SCC(fa);
			TreeSet<Integer> T=new TreeSet<Integer>();
			Iterator<FAState> s_it=s.getResult().iterator();
			while(s_it.hasNext()){
				T.add(st.getKey(s_it.next()));
			}
			int TailSize=0;
			TreeSet<Arc> isolatedArcs=h;
			while(TailSize!=T.size()){
				TailSize=T.size();
				TreeSet<Arc> isolatedArcsTemp=new TreeSet<Arc>();
				Iterator<Arc> arc_it=isolatedArcs.iterator();
				while(arc_it.hasNext()){
					Arc arc=arc_it.next();
					if(!T.contains(arc.getTo())){
						isolatedArcsTemp.add(arc);
					}else{
						T.add(arc.getFrom());
					}
				}
				isolatedArcs=isolatedArcsTemp;
			}
			Tail.put(h.toString(), T);
		}
		TreeSet<Integer> intersection = new TreeSet<Integer>();
		intersection.addAll(Head.get(g.toString()));
		intersection.retainAll(Tail.get(h.toString()));
		
		if(debug){
			if(intersection.isEmpty()){
				debug("g_graph:"+g+", Head: "+Head.get(g.toString()));
				debug("h_graph:"+h+", Tail: "+Tail.get(h.toString()));
			}
		}
		
		return !intersection.isEmpty();
	}

	private boolean lasso_finding_test_plus(TreeSet<Arc> g, TreeSet<Arc> h, int init){
		if(!Head.containsKey(g.toString())){
			TreeSet<Integer> H=new TreeSet<Integer>();
			Iterator<Arc> arc_g_it=g.iterator();
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				if(arc_g.getFrom()==init){
					H.add(arc_g.getTo());
					Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
					while(brel_it.hasNext()){
						Pair<FAState,FAState> p=brel_it.next();
						if(p.getLeft().getowner()==spec&&p.getRight().getowner()==spec){
							if(arc_g.getTo()==p.getLeft().id){
								H.add(p.getRight().id);
							}
						}
					}

				}
			}
			Head.put(g.toString(), H);
		}
		if(!Tail.containsKey(h.toString())){
			FiniteAutomaton fa=new FiniteAutomaton();
			OneToOneTreeMap<Integer,automata.FAState> st=new OneToOneTreeMap<Integer,automata.FAState>();
			Iterator<Arc> arc_h_it=h.iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				if(!st.containsKey(arc_h.getFrom())){
					st.put(arc_h.getFrom(), fa.createState());
				}
				if(!st.containsKey(arc_h.getTo())){
					st.put(arc_h.getTo(), fa.createState());
				}
				fa.addTransition(st.getValue(arc_h.getFrom()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");

				Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
				while(brel_it.hasNext()){
					Pair<FAState,FAState> p=brel_it.next();
					if(p.getLeft().getowner()==spec&&p.getRight().getowner()==spec){
						if(!st.containsKey(p.getLeft().getID()))
							st.put(p.getLeft().getID(), fa.createState());
						if(!st.containsKey(p.getRight().getID()))
							st.put(p.getRight().getID(), fa.createState());
						fa.addTransition(st.getValue(p.getLeft().getID()), st.getValue(p.getRight().getID()), "0");
					}
				}
			}
			SCC s=new SCC(fa);
			TreeSet<FAState> Pre=new TreeSet<FAState>();
			TreeSet<FAState> Next=s.getResult();
			Pre.addAll(Next);
			while(!Next.isEmpty()){
				FAState state=Next.iterator().next();
				Next.remove(state);
				TreeSet<FAState> toAdd=new TreeSet<FAState>();
				toAdd.addAll(state.getPre());
				toAdd.removeAll(Pre);
				Pre.addAll(toAdd);
				Next.addAll(toAdd);
			}
			Next=null;
			Iterator<FAState> Pre_it=Pre.iterator();
			TreeSet<Integer> T=new TreeSet<Integer>();
			while(Pre_it.hasNext()){
				FAState state=Pre_it.next();
				T.add(st.getKey(state)%spec.states.size());
			}
			Tail.put(h.toString(), T);
		}
		TreeSet<Integer> intersection = new TreeSet<Integer>();
		intersection.addAll(Head.get(g.toString()));
		intersection.retainAll(Tail.get(h.toString()));
		
		if(debug){
			if(intersection.isEmpty()){
				debug("g_graph:"+g+", Head: "+Head.get(g.toString()));
				debug("h_graph:"+h+", Tail: "+Tail.get(h.toString()));
			}
		}
		
		return !intersection.isEmpty();
	}
/*	private boolean lasso_finding_test_plus_old(TreeSet<Arc> g, TreeSet<Arc> h, int init){
		if(!Head.containsKey(g.toString())){
			TreeSet<Integer> H=new TreeSet<Integer>();
			Iterator<Arc> arc_g_it=g.iterator();
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				if(arc_g.getFrom()==init){
					H.add(arc_g.getTo());
				}
			}
			Head.put(g.toString(), H);
		}
		if(!Tail.containsKey(h.toString())){
			FiniteAutomaton fa=new FiniteAutomaton();
			OneToOneTreeMap<Integer,cav2010.automata.FAState> st=new OneToOneTreeMap<Integer,cav2010.automata.FAState>();
			Iterator<Arc> arc_h_it=h.iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				if(!st.containsKey(arc_h.getFrom())){
					st.put(arc_h.getFrom(), fa.createState());
					st.put(arc_h.getFrom()+spec.states.size(), fa.createState());
				}
				if(!st.containsKey(arc_h.getTo())){
					st.put(arc_h.getTo(), fa.createState());
					st.put(arc_h.getTo()+spec.states.size(), fa.createState());
				}
				fa.addTransition(st.getValue(arc_h.getFrom()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");
				fa.addTransition(st.getValue(arc_h.getFrom()+spec.states.size()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");

				Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
				while(brel_it.hasNext()){
					Pair<FAState,FAState> p=brel_it.next();
					if(p.getLeft().getowner()==spec&&p.getRight().getowner()==spec){
						if(!st.containsKey(p.getLeft().getID()))
							st.put(p.getLeft().getID(), fa.createState());
						if(!st.containsKey(p.getRight().getID()+spec.states.size()))
							st.put(p.getRight().getID()+spec.states.size(), fa.createState());
						fa.addTransition(st.getValue(p.getLeft().getID()), st.getValue(p.getRight().getID()+spec.states.size()), "0");
					}
				}
			}
			SCC s=new SCC(fa);
			TreeSet<FAState> Pre=new TreeSet<FAState>();
			TreeSet<FAState> Next=s.getResult();
			Pre.addAll(Next);
			while(!Next.isEmpty()){
				FAState state=Next.iterator().next();
				Next.remove(state);
				TreeSet<FAState> toAdd=new TreeSet<FAState>();
				toAdd.addAll(state.getPre());
				toAdd.removeAll(Pre);
				Pre.addAll(toAdd);
				Next.addAll(toAdd);
			}
			Next=null;
			Iterator<FAState> Pre_it=Pre.iterator();
			TreeSet<Integer> T=new TreeSet<Integer>();
			while(Pre_it.hasNext()){
				FAState state=Pre_it.next();
				T.add(st.getKey(state)%spec.states.size());
			}
			Tail.put(h.toString(), T);
		}
		TreeSet<Integer> intersection = new TreeSet<Integer>();
		intersection.addAll(Head.get(g.toString()));
		intersection.retainAll(Tail.get(h.toString()));
		
		if(debug){
			if(intersection.isEmpty()){
				debug("g_graph:"+g+", Head: "+Head.get(g.toString()));
				debug("h_graph:"+h+", Tail: "+Tail.get(h.toString()));
			}
		}
		
		return !intersection.isEmpty();
	}*/
	
	private Pair<TreeSet<Arc>, TreeSet<Arc>> min(Pair<TreeSet<Arc>, TreeSet<Arc>> metagraph){
		debug("before min:"+metagraph);
		TreeSet<Arc> left_result=new TreeSet<Arc>();
		Iterator<Arc> arc_it1 =metagraph.getLeft().iterator();
		while(arc_it1.hasNext()){
			Arc cur=arc_it1.next();
			boolean canAdd=true;
			Iterator<Arc> arc_it2 =metagraph.getLeft().iterator();
			while(arc_it2.hasNext()){
				Arc other=arc_it2.next();
				if(cur.getFrom()==other.getFrom()){
					if(!cur.getLabel()||other.getLabel()){
						if(cur.getTo()!=other.getTo()){
							if(frel.contains(new Pair<FAState,FAState>(new FAState(cur.getTo(),system),new FAState(other.getTo(), system )))){
								if(!frel.contains(new Pair<FAState,FAState>(new FAState(other.getTo(),system),new FAState(cur.getTo(), system )))){
									canAdd=false;
									break;
								}else if(cur.getTo()<other.getTo()){
									canAdd=false;
									break;
								}
							}
						}
					}
				}
			}			
			if(canAdd){
				left_result.add(cur);
			}
		}
		TreeSet<Arc> right_result=new TreeSet<Arc>();
		arc_it1 =metagraph.getRight().iterator();
		while(arc_it1.hasNext()){
			Arc cur=arc_it1.next();
			boolean canAdd=true;
			Iterator<Arc> arc_it2 =metagraph.getRight().iterator();
			while(arc_it2.hasNext()){
				Arc other=arc_it2.next();
				if(cur.getFrom()==other.getFrom()){
					if(!cur.getLabel()||other.getLabel()){
						if(cur.getTo()!=other.getTo()){
							if(frel.contains(new Pair<FAState,FAState>(new FAState(cur.getTo(),spec),new FAState(other.getTo(), spec )))){
								if(!frel.contains(new Pair<FAState,FAState>(new FAState(other.getTo(),spec),new FAState(cur.getTo(), spec )))){
									canAdd=false;
									break;
								}else if(cur.getTo()<other.getTo()){
									canAdd=false;
									break;
								}
							}
						}
					}
				}
			}			
			if(canAdd){
				right_result.add(cur);
			}
		}

		debug("after min:"+new Pair<TreeSet<Arc>, TreeSet<Arc>>(left_result,right_result));
		if((C1||C2)&&metagraph.getRight().size()!=right_result.size()){
			L.put(right_result, new TreeSet<Arc>());
			R.put(right_result, new TreeSet<Arc>());
			L.get(right_result).addAll(L.get(metagraph.getRight()));
			R.get(right_result).addAll(R.get(metagraph.getRight()));
			L.remove(metagraph.getRight());
			R.remove(metagraph.getRight());
		}
		return new Pair<TreeSet<Arc>, TreeSet<Arc>>(left_result,right_result);
	}

	private boolean contains(Set<Pair<FAState,FAState>> rel, FAState s1, FAState s2){
		return rel.contains(new Pair<FAState,FAState>(s1,s2));
	}
	private Pair<TreeSet<Arc>, TreeSet<Arc>> min_plus(Pair<TreeSet<Arc>, TreeSet<Arc>> metagraph){
		debug("before min:"+metagraph);
		TreeSet<Arc> left_result=new TreeSet<Arc>();
		Iterator<Arc> arc_it1 =metagraph.getLeft().iterator();
		while(arc_it1.hasNext()){
			Arc cur=arc_it1.next();
			boolean canAdd=true;
			Iterator<Arc> arc_it2 =metagraph.getLeft().iterator();
			while(arc_it2.hasNext()){
				Arc other=arc_it2.next();
				if(cur.getFrom()==other.getFrom()){
					if(!cur.getLabel()||other.getLabel()){
						if(cur.getTo()!=other.getTo()){
							if(frel.contains(new Pair<FAState,FAState>(new FAState(cur.getTo(),system),new FAState(other.getTo(), system )))){
								if(!frel.contains(new Pair<FAState,FAState>(new FAState(other.getTo(),system),new FAState(cur.getTo(), system )))){
									canAdd=false;
									break;
								}else if(cur.getTo()<other.getTo()){
									canAdd=false;
									break;
								}
							}
						}
					}
				}
			}			
			if(canAdd){
				left_result.add(cur);
			}
		}

		
		TreeSet<Arc> right_result=new TreeSet<Arc>();
		arc_it1 =metagraph.getRight().iterator();
		while(arc_it1.hasNext()){
			Arc cur=arc_it1.next();
			boolean canAdd=true;
			Iterator<Arc> arc_it2 =metagraph.getRight().iterator();
			while(arc_it2.hasNext()){
				Arc other=arc_it2.next();
				if(cur.getFrom()==other.getFrom() && cur.getTo()==other.getTo())
					continue;
				if(contains(brel, new FAState(cur.getFrom(),spec),new FAState(other.getFrom(), spec )) &&
				   (other.getFrom()>=cur.getFrom()||!contains(brel, new FAState(other.getFrom(),spec),new FAState(cur.getFrom(), spec )))){
					if(!cur.getLabel()||other.getLabel()){
						if(contains(frel, new FAState(cur.getTo(),spec),new FAState(other.getTo(), spec )) &&
						   (other.getTo()>=cur.getTo()||!contains(frel, new FAState(other.getTo(),spec),new FAState(cur.getTo(), spec )))){
									canAdd=false;
									break;
						}
					}
				}
			}
			if(canAdd){
				right_result.add(cur);
			}
		}
		debug("after min:"+new Pair<TreeSet<Arc>, TreeSet<Arc>>(left_result,right_result));
		if((C1||C2)&&metagraph.getRight().size()!=right_result.size()){
			L.put(right_result, new TreeSet<Arc>());
			R.put(right_result, new TreeSet<Arc>());
			L.get(right_result).addAll(L.get(metagraph.getRight()));
			R.get(right_result).addAll(R.get(metagraph.getRight()));
			L.remove(metagraph.getRight());
			R.remove(metagraph.getRight());
		}
		return new Pair<TreeSet<Arc>, TreeSet<Arc>>(left_result,right_result);
	}
	

	private ArrayList<Pair<TreeSet<Arc>, TreeSet<Arc>>> buildSingleCharacterMetaGraphs(){
		TreeMap<Pair<Arc, TreeSet<Arc>>, String> SGStringmap=new TreeMap<Pair<Arc, TreeSet<Arc>>, String>(new SuperGraphComparator());
		ArrayList<Pair<Arc, TreeSet<Arc>>> supergraphs=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
		Iterator<String> symbol_it=system.getAllTransitionSymbols().iterator();
		while(symbol_it.hasNext()){
			TreeSet<Arc> graph=new TreeSet<Arc>();
			String sym=symbol_it.next();
			Iterator<FAState> from_it=spec.states.iterator();
			while(from_it.hasNext()){
				FAState from=from_it.next();
				if(from.getNext(sym)!=null){
					Iterator<FAState> to_it=from.getNext(sym).iterator();
					while(to_it.hasNext()){
						FAState to=to_it.next();
						if(spec.F.contains(from)||spec.F.contains(to)){
							graph.add(new Arc(from.id,true,to.id));
						}else{
							graph.add(new Arc(from.id,false,to.id));
						}
					}
				}
			}

			from_it=system.states.iterator();
			while(from_it.hasNext()){
				FAState from=from_it.next();
				if(from.getNext(sym)!=null){
					Iterator<FAState> to_it=from.getNext(sym).iterator();
					while(to_it.hasNext()){
						FAState to=to_it.next();
						Arc left_arc;
						if(system.F.contains(from)||system.F.contains(to)){
							left_arc=new Arc(from.id,true,to.id);
						}else{
							left_arc=new Arc(from.id,false,to.id);
						}
						Pair<Arc, TreeSet<Arc>> supergraph=new Pair<Arc, TreeSet<Arc>>(left_arc,graph);
						ArrayList<Pair<Arc, TreeSet<Arc>>> toRemove=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
						boolean canAdd=true;
						Iterator<Pair<Arc, TreeSet<Arc>>> old_it=supergraphs.iterator();
						while(old_it.hasNext()){
							Pair<Arc, TreeSet<Arc>> old=old_it.next();
							if(smallerThan(old, supergraph)){
								canAdd=false;
								break;
							}else if(smallerThan(supergraph, old)){
								if(debug)
									stringmap.remove(old);
								toRemove.add(old);
							}
						}
						if(canAdd){
							if(opt2)
								if(!backward)
									supergraph=smin(supergraph);
								else
									supergraph=smin_plus(supergraph);

							supergraphs.add(supergraph);
							supergraphs.removeAll(toRemove);
							if(debug)
								SGStringmap.put(supergraph, sym);
						}
					}
				}
			}
		}
		HashMap<String, Pair<TreeSet<Arc>, TreeSet<Arc>>> MGmap=new HashMap<String, Pair<TreeSet<Arc>, TreeSet<Arc>>>();
		ArrayList<Pair<TreeSet<Arc>, TreeSet<Arc>>> metagraphs=new ArrayList<Pair<TreeSet<Arc>, TreeSet<Arc>>>();
		Iterator<Pair<Arc, TreeSet<Arc>>> SG_it=supergraphs.iterator();
		while(SG_it.hasNext()){
			Pair<Arc, TreeSet<Arc>> SG=SG_it.next();
			if(MGmap.containsKey(SG.getRight().toString())){
				MGmap.get(SG.getRight().toString()).getLeft().add(SG.getLeft());
			}else{
				TreeSet<Arc> left=new TreeSet<Arc>();
				left.add(SG.getLeft());
				Pair<TreeSet<Arc>, TreeSet<Arc>> MG=new Pair<TreeSet<Arc>, TreeSet<Arc>>(left, SG.getRight());
				MGmap.put(MG.getRight().toString(), MG);
				if(debug){
					stringmap.put(MG, SGStringmap.get(SG));
				}
				metagraphs.add(MG);
			}
			
		}

		if(C1||C2){
			Iterator<Pair<TreeSet<Arc>, TreeSet<Arc>>> Q1_it=metagraphs.iterator();
			while(Q1_it.hasNext()){
				Pair<TreeSet<Arc>, TreeSet<Arc>> g=Q1_it.next();
					Iterator<Arc> left_it=g.getLeft().iterator();
					while(left_it.hasNext()){
						Arc left=left_it.next();
						boolean c1=C1(left,g.getRight());
						boolean c2=C2(left,g.getRight());
						String label="";
						if(!c1&&!c2){
							addLabel("L", left, g.getRight());
							addLabel("R", left, g.getRight());
							label="RL";
						}
						else if(!c2){
							addLabel("R", left, g.getRight());
							label="R";
						}
						else if(!c1){
							addLabel("L", left, g.getRight());
							label+="L";
						}
					}
				}
		}
		
		
		return metagraphs;
	}
	
	//used only by initializing metagraphs
	private Pair<Arc, TreeSet<Arc>> smin(Pair<Arc, TreeSet<Arc>> supergraph){
		debug("before min:"+supergraph);
		TreeSet<Arc> result=new TreeSet<Arc>();
		Iterator<Arc> arc_it1 =supergraph.getRight().iterator();
		while(arc_it1.hasNext()){
			Arc cur=arc_it1.next();
			boolean canAdd=true;
			Iterator<Arc> arc_it2 =supergraph.getRight().iterator();
			while(arc_it2.hasNext()){
				Arc other=arc_it2.next();
				if(cur.getFrom()==other.getFrom()){
					if(!cur.getLabel()||other.getLabel()){
						if(cur.getTo()!=other.getTo()){
							if(frel.contains(new Pair<FAState,FAState>(new FAState(cur.getTo(),spec),new FAState(other.getTo(), spec )))){
								if(!frel.contains(new Pair<FAState,FAState>(new FAState(other.getTo(),spec),new FAState(cur.getTo(), spec )))){
									canAdd=false;
									break;
								}else if(cur.getTo()<other.getTo()){
									canAdd=false;
									break;
								}
							}
						}
					}
				}
			}			
			if(canAdd){
				result.add(cur);
			}
		}
		debug("after min:"+new Pair<Arc, TreeSet<Arc>>(supergraph.getLeft(),result));
		return new Pair<Arc, TreeSet<Arc>>(supergraph.getLeft(),result);
	}

	//used only by initializing metagraphs
	private Pair<Arc, TreeSet<Arc>> smin_plus(Pair<Arc, TreeSet<Arc>> supergraph){
		debug("before min:"+supergraph);
		TreeSet<Arc> result=new TreeSet<Arc>();
		Iterator<Arc> arc_it1 =supergraph.getRight().iterator();
		while(arc_it1.hasNext()){
			Arc cur=arc_it1.next();
			boolean canAdd=true;
			Iterator<Arc> arc_it2 =supergraph.getRight().iterator();
			while(arc_it2.hasNext()){
				Arc other=arc_it2.next();
				if(cur.getFrom()==other.getFrom() && cur.getTo()==other.getTo())
					continue;
				if(contains(brel, new FAState(cur.getFrom(),spec),new FAState(other.getFrom(), spec )) &&
				   (other.getFrom()>=cur.getFrom()||!contains(brel, new FAState(other.getFrom(),spec),new FAState(cur.getFrom(), spec )))){
					if(!cur.getLabel()||other.getLabel()){
						if(contains(frel, new FAState(cur.getTo(),spec),new FAState(other.getTo(), spec )) &&
						   (other.getTo()>=cur.getTo()||!contains(frel, new FAState(other.getTo(),spec),new FAState(cur.getTo(), spec )))){
									canAdd=false;
									break;
						}
					}
				}
			}
			if(canAdd){
				result.add(cur);
			}
		}
		debug("after min:"+new Pair<Arc, TreeSet<Arc>>(supergraph.getLeft(),result));
		return new Pair<Arc, TreeSet<Arc>>(supergraph.getLeft(),result);
	}
	

	private Pair<TreeSet<Arc>, TreeSet<Arc>> compose(Pair<TreeSet<Arc>, TreeSet<Arc>> g, Pair<TreeSet<Arc>, TreeSet<Arc>> h, TreeMap<TreeSet<Arc>, TreeSet<Arc>> oriL){
		TreeSet<Arc> right=new TreeSet<Arc>();
		Iterator<Arc> arc_g_it=g.getRight().iterator();
		while(arc_g_it.hasNext()){
			Arc arc_g=arc_g_it.next();
			Iterator<Arc> arc_h_it=h.getRight().iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				boolean connected=false;
				if(backward &&contains(brel, new FAState(arc_g.getTo(),spec),new FAState(arc_h.getFrom(), spec )))
					connected=true;
				else if(!backward&&arc_g.getTo()==arc_h.getFrom())
					connected=true;
				if(connected){
					if(arc_g.getLabel()||arc_h.getLabel()){
						right.add(new Arc(arc_g.getFrom(),true,arc_h.getTo()));
						right.remove(new Arc(arc_g.getFrom(),false,arc_h.getTo()));
					}else{
						if(!right.contains(new Arc(arc_g.getFrom(),true,arc_h.getTo()))){
							right.add(new Arc(arc_g.getFrom(),false,arc_h.getTo()));
						}
					}
				}				
			}			
		}
		
		if((C1||C2) && !L.containsKey(right))
			L.put(right, new TreeSet<Arc>());
		if((C1||C2) && !R.containsKey(right))
			R.put(right, new TreeSet<Arc>());
		
		TreeSet<Arc> left=new TreeSet<Arc>();
		Iterator<Arc> arc_i_it=g.getLeft().iterator();
		while(arc_i_it.hasNext()){
			Arc arc_i=arc_i_it.next();
			short label=0;
			if(C1||C2){
				if(oriL.get(g.getRight())!=null && oriL.get(g.getRight()).contains(arc_i)){
					if(R.get(g.getRight()) !=null && R.get(g.getRight()).contains(arc_i)){
						label=2;
						debug("Left arc is LR labeled");
						//LR labeled (2)
					}else{
						label=1;
						debug("Left arc is L labeled");
						//L labeled (1)
					}
				}else{
					//else R labeled (0)
					debug("Left arc is R labeled");

				}
			}
			Iterator<Arc> arc_h_it=h.getLeft().iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				if(arc_i.getTo()==arc_h.getFrom()){
					if(arc_i.getLabel()||arc_h.getLabel()){
						Arc newarc=new Arc(arc_i.getFrom(),true,arc_h.getTo());
						left.add(newarc);
						left.remove(new Arc(arc_i.getFrom(),false,arc_h.getTo()));
						if(C1||C2){
							if(L.get(right)==null)
								L.put(right, new TreeSet<Arc>());
							if(R.get(right)==null)
								R.put(right, new TreeSet<Arc>());
							L.get(right).remove(new Arc(arc_i.getFrom(),false,arc_h.getTo()));
							R.get(right).remove(new Arc(arc_i.getFrom(),false,arc_h.getTo()));
							if(label==0 && !C2(newarc, right)){
								R.get(right).add(newarc);
								debug("Add R label to "+newarc);
							}else if(label==1 && !C1(newarc, right)){
								L.get(right).add(newarc);
								debug("Add L label to "+newarc);
							}else{
								if(!C2(newarc, right)){
									R.get(right).add(newarc);
									debug("Add R label to "+newarc);
								}
								if(!C1(newarc, right)){
									L.get(right).add(newarc);
									debug("Add L label to "+newarc);
								}
							}
						}
					}else{
						if(!left.contains(new Arc(arc_i.getFrom(),true,arc_h.getTo()))){
							Arc newarc=new Arc(arc_i.getFrom(),false,arc_h.getTo());
							left.add(new Arc(arc_i.getFrom(),false,arc_h.getTo()));
							if(C1||C2){
								if(label==0 && !C2(newarc, right)){
									R.get(right).add(newarc);
								}else if(label==1 && !C1(newarc, right)){
									L.get(right).add(newarc);
								}else{
									if(!C2(newarc, right))
										R.get(right).add(newarc);
									if(!C1(newarc, right))
										L.get(right).add(newarc);
								}
							}
						}
					}
				}				
			}			
		}

		
		if(left.size()==0){
			return null;
		}else{
			return new Pair<TreeSet<Arc>, TreeSet<Arc>>(left,right);
		}
	}
	//clean metagraph src w.r.t tgt
	//if tgt.right not smaller than src.right then does nothing
	boolean clean(Pair<TreeSet<Arc>, TreeSet<Arc>> src, Pair<TreeSet<Arc>, TreeSet<Arc>> tgt, short type){
		
		Iterator<Arc> arc_g_it=tgt.getRight().iterator();
		while(arc_g_it.hasNext()){
			Arc arc_g=arc_g_it.next();
			boolean has_larger=false;
			Iterator<Arc> arc_h_it=src.getRight().iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				boolean fromOkay=false;
				if(backward &&brel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getFrom(),spec),new FAState(arc_h.getFrom(), spec ))))
					fromOkay=true;
				else if(!backward&&arc_g.getFrom()==arc_h.getFrom())
					fromOkay=true;
				if(fromOkay){
					if(!arc_g.getLabel()||arc_h.getLabel()){
						if(frel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getTo(),spec),new FAState(arc_h.getTo(),spec)))){
							has_larger=true;
							break;
						}
					}
				}
			}			
			if(!has_larger){
				return true;
			}
		}
		debug("Before clean"+type+" :"+src+" (w.r.t. "+tgt+")");
		//tgt.right smaller than src.right

		if(type!=3 && src.getRight().size()==tgt.getRight().size()){
			arc_g_it=src.getRight().iterator();
			boolean has_larger=true;
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				has_larger=false;
				Iterator<Arc> arc_h_it=tgt.getRight().iterator();
				while(arc_h_it.hasNext()){
					Arc arc_h=arc_h_it.next();
					boolean fromOkay=false;
					if(backward &&brel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getFrom(),spec),new FAState(arc_h.getFrom(), spec ))))
						fromOkay=true;
					else if(!backward&&arc_g.getFrom()==arc_h.getFrom())
						fromOkay=true;
					if(fromOkay){
						if(!arc_g.getLabel()||arc_h.getLabel()){
							if(frel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getTo(),spec),new FAState(arc_h.getTo(),spec)))){
								has_larger=true;
								break;
							}
						}
					}
				}			
				if(!has_larger){
					break;
				}
			}
			if(has_larger){
				//src.right := tgt.right
				if(L.get(src.getRight())!=null){
					if(L.get(tgt.getRight())==null){
						L.put(tgt.getRight(), new TreeSet<Arc>());
					}
					L.get(tgt.getRight()).addAll(L.get(src.getRight()));
				}
				if(R.get(src.getRight())!=null){
					if(R.get(tgt.getRight())==null){
						R.put(tgt.getRight(), new TreeSet<Arc>());
					}
					R.get(tgt.getRight()).addAll(R.get(src.getRight()));
				}
				src.getRight().clear();
				src.getRight().addAll(tgt.getRight());
			}
		}

		boolean regain=false;
		TreeSet<Arc> left=new TreeSet<Arc>();
		arc_g_it=src.getLeft().iterator();
		while(arc_g_it.hasNext()){
			Arc arc_g=arc_g_it.next();
			boolean has_larger=false;
			Iterator<Arc> arc_h_it=tgt.getLeft().iterator();
			Arc arc_h = null;
			while(arc_h_it.hasNext()){
				arc_h=arc_h_it.next();
				if(arc_g.getFrom()==arc_h.getFrom()){
					if(!arc_g.getLabel()||arc_h.getLabel()){
						if(frel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getTo(),system),new FAState(arc_h.getTo(),system)))){
							has_larger=true;
							break;
						}
					}
				}
			}			
			if(!has_larger){
				left.add(arc_g);
			}else if(type==2){
				if(L.get(src.getRight())!=null && L.get(src.getRight()).contains(arc_g)){
					if(L.get(tgt.getRight())==null)
						L.put(tgt.getRight(), new TreeSet<Arc>());
					if(!L.get(tgt.getRight()).contains(arc_h)){
						L.get(tgt.getRight()).add(arc_h);
						regain=true;
					}
				}
			}
		}
		
		
		
		
//		if(src.toString().compareTo(tgt.toString())==0&& left.size()!=0)
//			System.out.println();
		src.getLeft().clear();
		src.getLeft().addAll(left);
		debug("After clean"+type+" :"+src);
		if(regain)
			debug("Regain L :"+tgt);
		return !regain;
	
	}
	
	boolean smallerThan(Pair<Arc, TreeSet<Arc>> old, Pair<Arc, TreeSet<Arc>> old2){
		Arc old_arc=old.getLeft();
		Arc old2_arc=old2.getLeft();
		if(!(old_arc.getFrom()==old2_arc.getFrom() && (old_arc.getLabel() ||!old2_arc.getLabel()) &&
				frel.contains(new Pair<FAState,FAState>(new FAState(old2_arc.getTo(),system),new FAState(old_arc.getTo(),system))))){
			return false;
		}
		
		Iterator<Arc> arc_g_it=old.getRight().iterator();
		while(arc_g_it.hasNext()){
			Arc arc_g=arc_g_it.next();
			boolean has_larger=false;
			Iterator<Arc> arc_h_it=old2.getRight().iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				boolean fromOkay=false;
				if(backward &&brel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getFrom(),spec),new FAState(arc_h.getFrom(), spec ))))
					fromOkay=true;
				else if(!backward&&arc_g.getFrom()==arc_h.getFrom())
					fromOkay=true;
				if(fromOkay){
					if(!arc_g.getLabel()||arc_h.getLabel()){
						if(frel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getTo(),spec),new FAState(arc_h.getTo(),spec)))){
							has_larger=true;
							break;
						}
					}
				}
			}			
			if(!has_larger){
				return false;
			}
		}
		return true;
	}
	public void inclusionTest(){
		this.computeSim();

		debug("Aut A:");
		debug(system.toString());
		debug("Aut B:");
		debug(spec.toString());
		debug("FSim="+showSim(frel));
		if(backward)
		debug("BSim="+showSim(brel));
		this.included=included();
	}
	public long getCpuTime( ) {
	    ThreadMXBean bean = ManagementFactory.getThreadMXBean( );
	    return bean.isCurrentThreadCpuTimeSupported( ) ?
	        bean.getCurrentThreadCpuTime( ) : 0L;
	}
	
	public boolean isIncluded(){
		return included;
	}
	@Override
	public void run(){

		runTime=getCpuTime();
			inclusionTest();
		runTime=getCpuTime()-runTime;
	}
	
	public long getRunTime(){
		return runTime;
	}


	void computeSim(){

		ArrayList<FAState> all_states=new ArrayList<FAState>();
		HashSet<String> alphabet=new HashSet<String>();
		
		all_states.addAll(spec.states);
		alphabet.addAll(spec.alphabet);

		all_states.addAll(system.states);
		alphabet.addAll(system.alphabet);
		
		
		FAState[] states=all_states.toArray(new FAState[0]);		
		boolean[] isFinal = new boolean[states.length];
		boolean[] isInit = new boolean[states.length];
		boolean[][] fsim = new boolean[states.length][states.length];
		boolean[][] bsim = new boolean[states.length][states.length];
		// sim[u][v]=true iff v in sim(u) iff v simulates u
		for(int i=0;i<states.length;i++){			
			isFinal[i] = states[i].getowner().F.contains(states[i]);
			isInit[i] =states[i].getowner().getInitialState().compareTo(states[i])==0;
		}
		for(int i=0;i<states.length;i++){
			for(int j=i;j<states.length;j++){
				fsim[i][j] = (!isFinal[i] || isFinal[j]) && states[j].fw_covers(states[i]);
				fsim[j][i] = (isFinal[i] || !isFinal[j]) && states[i].fw_covers(states[j]);
				if(backward){
					bsim[i][j] = (!isInit[i] || isInit[j]) && (!isFinal[i] || isFinal[j]) && states[j].bw_covers(states[i]);
					bsim[j][i] = (isInit[i] || !isInit[j]) && (isFinal[i] || !isFinal[j]) && states[i].bw_covers(states[j]);
				}
			}
		}
		Simulation sim = new Simulation();
		frel = sim.FastFSimRelNBW(spec,system,fsim);
		if(quotient){
			spec=quotient(spec, frel);
			system=quotient(system, frel);
		}

		if(backward){
			all_states.clear();
			all_states.addAll(spec.states);
			all_states.addAll(system.states);
			states=all_states.toArray(new FAState[0]);		
			isFinal = new boolean[states.length];
			isInit = new boolean[states.length];
			bsim = new boolean[states.length][states.length];
			for(int i=0;i<states.length;i++){			
				isFinal[i] = states[i].getowner().F.contains(states[i]);
				isInit[i] =states[i].getowner().getInitialState().compareTo(states[i])==0;
			}
			for(int i=0;i<states.length;i++){
				for(int j=i;j<states.length;j++){
					bsim[i][j] = (!isInit[i] || isInit[j]) && (!isFinal[i] || isFinal[j]) && states[j].bw_covers(states[i]);
					bsim[j][i] = (isInit[i] || !isInit[j]) && (isFinal[i] || !isFinal[j]) && states[i].bw_covers(states[j]);
				}
			}
			brel = sim.FastBSimRelNBW(spec,system,bsim);
		}

	}	
	
	TreeMap<Pair<TreeSet<Arc>, TreeSet<Arc>>, String> stringmap=new TreeMap<Pair<TreeSet<Arc>, TreeSet<Arc>>, String>(new MetaGraphComparator());
	static 	public int mggen=0;

	private boolean included(){
		ArrayList<Pair<TreeSet<Arc>, TreeSet<Arc>>> Q1=this.buildSingleCharacterMetaGraphs();
		debug("Init:"+Q1);
		
		Iterator<Pair<TreeSet<Arc>, TreeSet<Arc>>> Q1_it=Q1.iterator();

		while(Q1_it.hasNext()){
			Pair<TreeSet<Arc>, TreeSet<Arc>> g=Q1_it.next();
			if(!double_graph_test(g, g, spec.getInitialState().id)){
				return false;
			}
		}

		
		TreeSet<Pair<TreeSet<Arc>, TreeSet<Arc>>> Next=new TreeSet<Pair<TreeSet<Arc>, TreeSet<Arc>>>(new MetaGraphComparator());
		TreeSet<Pair<TreeSet<Arc>, TreeSet<Arc>>> Processed=new TreeSet<Pair<TreeSet<Arc>, TreeSet<Arc>>>(new MetaGraphComparator());
		Next.addAll(Q1);
		mggen+=Q1.size();
		while(!Next.isEmpty()&&(!C1||!L.isEmpty())){
			if(stop)
				break;
			debug("");
			debug("Begin while loop");
			debug("Processed:"+Processed);
			debug("Next:"+Next);
			
			Pair<TreeSet<Arc>, TreeSet<Arc>> g=Next.first();
			debug("Pick "+g+"from Next");
			
			Iterator<Pair<TreeSet<Arc>, TreeSet<Arc>>> Processed_it=Processed.iterator();
			while(Processed_it.hasNext()){
				Pair<TreeSet<Arc>, TreeSet<Arc>> h=Processed_it.next();
				if(!double_graph_test(g, h, spec.getInitialState().id)||!double_graph_test(h, g, spec.getInitialState().id))
					return false;
			}
			Next.remove(g);
//			boolean merged=false;
//			Processed_it=Processed.iterator();
//			while(Processed_it.hasNext()){
//				Pair<TreeSet<Arc>, TreeSet<Arc>> cur=Processed_it.next();
//				if(cur.getRight().toString().compareToIgnoreCase(g.getRight().toString())==0){
//					merged=true;
//					cur.getLeft().addAll(g.getLeft());
//					break;
//				}
//			}
//			if(!merged)
				Processed.add(g);

			TreeMap<TreeSet<Arc>,TreeSet<Arc>> ori_L=new TreeMap<TreeSet<Arc>,TreeSet<Arc>>(new GraphComparator());
			if(C1||C2){
				Iterator<TreeSet<Arc>> L_it=L.keySet().iterator();
				while(L_it.hasNext()){
					TreeSet<Arc> key=new TreeSet<Arc>();
					key.addAll(L_it.next());
					TreeSet<Arc> obj=new TreeSet<Arc>();
					obj.addAll(L.get(key));
					ori_L.put(key, obj);
				}
//				ori_L.putAll(this.L);
				if(L.get(g.getRight())==null)
					L.put(g.getRight(), new TreeSet<Arc>());
				L.get(g.getRight()).removeAll(g.getLeft());
				debug("Move"+g+"from Next to Processed with L label removed");
			}else{
				debug("Move"+g+"from Next to Processed");
			}
			
			Q1_it=Q1.iterator();
			while(Q1_it.hasNext()){
				Pair<TreeSet<Arc>, TreeSet<Arc>> h=Q1_it.next();
				Pair<TreeSet<Arc>, TreeSet<Arc>> f=compose(g, h,ori_L);
				if(f==null)
					continue;
				if(!backward)
					f=min(f);
				else
					f=min_plus(f);
				if(debug)
					stringmap.put(f, stringmap.get(g)+stringmap.get(h));
				debug("Produce f:"+f+"("+stringmap.get(f) +")=Min_s("+g+"("+stringmap.get(g) +");"+h+"("+stringmap.get(h) +"))");

				ArrayList<Pair<TreeSet<Arc>, TreeSet<Arc>>> toRemove=new ArrayList<Pair<TreeSet<Arc>, TreeSet<Arc>>>();
				ArrayList<Pair<TreeSet<Arc>, TreeSet<Arc>>> toAdd=new ArrayList<Pair<TreeSet<Arc>, TreeSet<Arc>>>();
				Iterator<Pair<TreeSet<Arc>, TreeSet<Arc>>> Next_it=Next.iterator();
				while(Next_it.hasNext()){
					Pair<TreeSet<Arc>, TreeSet<Arc>> p=Next_it.next();
					String index="";
					if(debug){
						index=stringmap.get(f);
						stringmap.remove(f);
					}
					clean(f, p, (short) 1);
					if(debug){
						stringmap.put(f, index);
					}

					if(f.getLeft().size()==0)
						break;
					
					
					TreeSet<Arc> left=new TreeSet<Arc>();
					left.addAll(p.getLeft());
					TreeSet<Arc> right=new TreeSet<Arc>();
					right.addAll(p.getRight());
					Pair<TreeSet<Arc>, TreeSet<Arc>> p_new=new Pair<TreeSet<Arc>, TreeSet<Arc>>(left, right); 
					clean(p_new, f, (short) 3);
					
					if(p_new.toString().compareTo(p.toString())!=0){
						stringmap.put(p_new, stringmap.get(p));
						toRemove.add(p);
						if(p_new.getLeft().size()>0)
							toAdd.add(p_new);
					}
				}
				if(toAdd.size()>0){
					Next.addAll(toAdd);
					toAdd.clear();
				}
				if(f.getLeft().size()==0)
					continue;
				Processed_it=Processed.iterator();
				while(Processed_it.hasNext()){
					Pair<TreeSet<Arc>, TreeSet<Arc>> p=Processed_it.next();
					String index="";
					if(debug){
						index=stringmap.get(f);
						stringmap.remove(f);
					}
					if(!clean(f, p,(short) 2)&&(C1||C2)){//p regain L somewhere
						Iterator<Pair<TreeSet<Arc>, TreeSet<Arc>>> Proc_it=Processed.iterator();
						while(Proc_it.hasNext()){
							Pair<TreeSet<Arc>, TreeSet<Arc>> right_ext=Proc_it.next();
							if(!double_graph_test(f, right_ext, spec.getInitialState().id))
								return false;
						}
					}
					if(debug){
						stringmap.put(f, index);
					}
					if(f.getLeft().size()==0)
						break;
					TreeSet<Arc> left=new TreeSet<Arc>();
					left.addAll(p.getLeft());
					TreeSet<Arc> right=new TreeSet<Arc>();
					right.addAll(p.getRight());
					Pair<TreeSet<Arc>, TreeSet<Arc>> p_new=new Pair<TreeSet<Arc>, TreeSet<Arc>>(left, right); 
					clean(p_new, f, (short) 3);
					if(p_new.toString().compareTo(p.toString())!=0){
						stringmap.put(p_new, stringmap.get(p));
//						if(stringmap.get(p).length()>1)
//							stringmap.remove(p);
						toRemove.add(p);
						if(p_new.getLeft().size()>0)
							toAdd.add(p_new);
					}
				}
				if(f.getLeft().size()==0)
					continue;
				if(!double_graph_test(f, f, spec.getInitialState().id))
					return false;
				Processed.removeAll(toRemove);
				Next.removeAll(toRemove);
				Processed.addAll(toAdd);
				String index="";
				if(debug){
					index=stringmap.get(f);
					stringmap.remove(f);
				}
				if(debug){
					stringmap.put(f, index);
				}
				mggen++;
//				Next_it=Next.iterator();
//				while(Next_it.hasNext()){
//					Pair<TreeSet<Arc>, TreeSet<Arc>> cur=Next_it.next();
//					if(cur.getRight().toString().compareToIgnoreCase(f.getRight().toString())==0){
//						merged=true;
//						cur.getLeft().addAll(f.getLeft());
//						break;
//					}
//				}
//				if(!merged)
					Next.add(f);
				debug("Add "+f+" to Next");
				
			}
		}
		return true;
	}			

	
	
		
	
	private void addLabel(String type, Arc left,
			TreeSet<Arc> right) {
		TreeMap<TreeSet<Arc>, TreeSet<Arc>> Label=null;
		if(type.compareTo("L")==0)
			Label =L;
		else
			Label =R;
		debug("("+left+", "+right+") adds "+type+" label");
		
		if(Label.get(right)==null){
			TreeSet<Arc> empty=new TreeSet<Arc>();
			Label.put(right, empty);
		}
		Label.get(right).add(left);
	}
//	private void delLabel(TreeMap<TreeSet<Arc>, TreeSet<Arc>> Label, Arc left,
//			TreeSet<Arc> right) {
//		if(Label.get(right)==null){
//			return;
//		}
//		Label.get(right).remove(left);
//	}

	public void stopIt(){
		stop=true;
	}
}
