/* Written by Yu-Fang Chen, Richard Mayr, and Chih-Duo Hong               */
/* Copyright (c) 2010                  	                                  */
/*                                                                        */
/* This program is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation; either version 2 of the License, or      */
/* (at your option) any later version.                                    */
/*                                                                        */
/* This program is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */
/*                                                                        */
/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA*/

package algorithms;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;

import automata.FAState;
import automata.FiniteAutomaton;
import comparator.NewSuperGraphComparator;
import comparator.StatePairComparator;
import comparator.SuperGraphComparator;
import datastructure.Arc;
import datastructure.Graph;
import datastructure.HashSet;
import datastructure.Metagraph;
import datastructure.MetagraphBV;
import datastructure.NewMetagraph;
import datastructure.OneToOneTreeMap;
import datastructure.Pair;


/**
 * 
 * @author Yu-Fang Chen
 * 
 */
public class InclusionOptBVLayered{
	
	
	private int verbose =10;
	int verboseC1=1;
	int verboseMin=1;

	public int removedCnt;
	public boolean included=true;
	public int cntL=0;
		
	private TreeMap<Graph,BitSet> SysTail=new TreeMap<Graph,BitSet>();
	private TreeMap<Graph,BitSet> SysHead=new TreeMap<Graph,BitSet>();

	private TreeMap<Graph,BitSet> SpecTail=new TreeMap<Graph,BitSet>();
	private TreeMap<Graph,BitSet> SpecHead=new TreeMap<Graph,BitSet>();

	private long runTime;
	private boolean stop=false;
	private Set<Pair<FAState,FAState>> frel, brel;
	private Map<FAState, Set<FAState>> frelM, brelM;
	FiniteAutomaton spec, system;
	String showSim(Set<Pair<FAState,FAState>> sim){
		String result="";
		Iterator<Pair<FAState,FAState>> sim_it=sim.iterator();
		while(sim_it.hasNext()){
			Pair<FAState,FAState> p=sim_it.next();
			if(p.getLeft().compareTo(p.getRight())!=0)
				result+=("("+p.getLeft()+","+p.getRight()+")");
		}
		return result;
	}
	void debug(String out, int v){
		if(Options.debug&&v>verbose)
		System.out.println(out);
	}
	
	
	public InclusionOptBVLayered(FiniteAutomaton system, FiniteAutomaton spec){
		this.spec=spec;
		this.system=system;
	}
	/**
	 * Simplify a finite automaton by merging simulation equivalent states
	 * @param fa: a finite automaton
	 * @param Sim: some simulation rel_specation over states in the spec automaton
	 * 
	 * @return an equivalent finite automaton
	 */
	private FiniteAutomaton quotient(FiniteAutomaton fa, Set<Pair<FAState,FAState>> rel) {
		FiniteAutomaton result=new FiniteAutomaton();
		result.name=fa.name;
		TreeMap<FAState,FAState> map=new TreeMap<FAState,FAState>();
		TreeMap<FAState,FAState> reducedMap=new TreeMap<FAState,FAState>();
		
		Iterator<FAState> state_it=fa.states.iterator();
		while(state_it.hasNext()){
			FAState state=state_it.next();
			map.put(state, state);
			Iterator<FAState> state_it2=fa.states.iterator();
			while(state_it2.hasNext()){
				FAState state2=state_it2.next();
				if(rel.contains(new Pair<FAState,FAState>(state,state2)) &&
					rel.contains(new Pair<FAState,FAState>(state2,state))){
					map.put(state,state2);
				}
			}			
		}

		FAState init=result.createState();
		reducedMap.put(map.get(fa.getInitialState()), init);
		result.setInitialState(init);

		state_it=fa.states.iterator();
		while(state_it.hasNext()){
			FAState state=state_it.next();
			if(!reducedMap.containsKey(map.get(state))){
				reducedMap.put(map.get(state), result.createState());
			}
			if(fa.F.contains(state)){
				result.F.add(reducedMap.get(map.get(state)));
			}
			Iterator<String> sym_it=state.nextIt();
			while(sym_it.hasNext()){
				String sym=sym_it.next();
				Iterator<FAState> to_it=state.getNext(sym).iterator();
				while(to_it.hasNext()){
					FAState to=to_it.next();
					if(!reducedMap.containsKey(map.get(to))){
						reducedMap.put(map.get(to), result.createState());
					}
					result.addTransition(reducedMap.get(map.get(state)), reducedMap.get(map.get(to)), sym);
				}
			}
		}
		Set<Pair<FAState,FAState>> newrel=new TreeSet<Pair<FAState,FAState>>(new StatePairComparator());
		Iterator<Pair<FAState,FAState>> sim_it=rel.iterator();
		while(sim_it.hasNext()){
			Pair<FAState,FAState> sim=sim_it.next();
			FAState left,right;
			if(sim.getLeft().getowner()==fa){
				left=reducedMap.get(map.get(sim.getLeft()));
			}else{
				left=sim.getLeft();
			}
			
			if(sim.getRight().getowner()==fa){
				right=reducedMap.get(map.get(sim.getRight()));
			}else{
				right=sim.getRight();
			}
			newrel.add(new Pair<FAState,FAState>(left,right));
		}
		rel.clear();
		rel.addAll(newrel);
		
		return result;
	}	

	private boolean C1(Arc left,Graph right){
		if(!Options.C1){
			return false;
		}
		if(left.getFrom()!=system.getInitialState().id){
			debug("C1("+left+","+right+")="+true,verboseC1);
			return true;
		}
		
		if(!Options.layered){
			Iterator<Arc> right_it=right.iterator();
			while(right_it.hasNext()){
				Arc r=right_it.next();
				if(r.getFrom()==spec.getInitialState().id){
					if(frelM.get(new FAState(left.getTo(),system)).contains(new FAState(r.getTo(),spec))){
						debug("C1("+left+","+right+")="+true,verboseC1);
						return true;
					}
				}
			}
		}else{
			Iterator<Integer> leftSt_it=right.leftStateIt();
			while(leftSt_it.hasNext()){
				Iterator<Arc> right_it=right.iterator(leftSt_it.next());
				while(right_it.hasNext()){
					Arc r=right_it.next();
					if(r.getFrom()==spec.getInitialState().id){
						if(frelM.get(new FAState(left.getTo(),system)).contains(new FAState(r.getTo(),spec))){
							debug("C1("+left+","+right+")="+true,verboseC1);
							return true;
						}
					}
				}
			}
		}
		debug("C1("+left+","+right+")="+false,verboseC1);
		return false;
	}

	
	private boolean double_graph_test(MetagraphBV hl, MetagraphBV gr){
		if(hl.getLeft().intersects(gr.getLeft())&&!hl.getRight().intersects(gr.getRight())){
//			System.out.println("System:"+hl.getLeft()+":"+gr.getLeft());
//			System.out.println("Spec:"+hl.getRight()+":"+gr.getRight());
			
			return false;
		}else
			return true;
	}	
		
	private BitSet getL(Graph g, int init, boolean isSpec/*spec takes upward*/){
		
		if(!(isSpec&& SpecHead.containsKey(g))&&!(!isSpec&& SysHead.containsKey(g))){
			BitSet H=new BitSet();

			if(!Options.layered){
				Iterator<Arc> arc_g_it=g.iterator();
				while(arc_g_it.hasNext()){
					Arc arc_g=arc_g_it.next();
					if((arc_g.L||isSpec/*arcs in spec do not have label*/)&&arc_g.getFrom()==init){
						H.set(arc_g.getTo());
					}
					
				}
			}else{
				Iterator<Integer> leftSt_it=g.leftStateIt();
				while(leftSt_it.hasNext()){
					Iterator<Arc> arc_g_it=g.iterator(leftSt_it.next());
					while(arc_g_it.hasNext()){
						Arc arc_g=arc_g_it.next();
						if((arc_g.L||isSpec/*arcs in spec do not have label*/)&&arc_g.getFrom()==init){
							H.set(arc_g.getTo());
						}
						
					}
				}
			}
			boolean changed=true;
			while(changed){
				changed=false;
				if(isSpec){
					Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
					while(brel_it.hasNext()){
						Pair<FAState,FAState> p=brel_it.next();
						if(p.getLeft().getowner()==spec&&p.getRight().getowner()==spec){
							if(H.get(p.getLeft().id)&&!H.get(p.getRight().id)){
								H.set(p.getRight().id);
								changed=true;
							}
						}
					}
				}else{
					Iterator<Pair<FAState,FAState>>frel_it=frel.iterator();
					while(frel_it.hasNext()){
						Pair<FAState,FAState> p=frel_it.next();
						if(p.getLeft().getowner()==system&&p.getRight().getowner()==system){
							if(H.get(p.getRight().id)&&!H.get(p.getLeft().id)){
								H.set(p.getLeft().id);
								changed=true;
							}
						}
					}
				}
			}
			if(isSpec)
				SpecHead.put(g.clone(), H);
			else
				SysHead.put(g.clone(), H);
		}
		if(isSpec)
			return SpecHead.get(g);
		else
			return SysHead.get(g);
	}	

	private BitSet getR(Graph h, boolean isSystem){
		if(!(isSystem&&SysTail.containsKey(h)&&!(!isSystem&&SpecTail.containsKey(h)))){
			FiniteAutomaton fa=new FiniteAutomaton();
			OneToOneTreeMap<Integer,automata.FAState> st=new OneToOneTreeMap<Integer,automata.FAState>();
			
			if(!Options.layered){
				Iterator<Arc> arc_h_it=h.iterator();
				while(arc_h_it.hasNext()){
					Arc arc_h=arc_h_it.next();
					if(!st.containsKey(arc_h.getFrom()))
						st.put(arc_h.getFrom(), fa.createState());
					if(!st.containsKey(arc_h.getTo()))
						st.put(arc_h.getTo(), fa.createState());
					fa.addTransition(st.getValue(arc_h.getFrom()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");
				}
			}else{
				Iterator<Integer> leftSt_it=h.leftStateIt();
				while(leftSt_it.hasNext()){
					Iterator<Arc> arc_h_it=h.iterator(leftSt_it.next());
					while(arc_h_it.hasNext()){
						Arc arc_h=arc_h_it.next();
						if(!st.containsKey(arc_h.getFrom()))
							st.put(arc_h.getFrom(), fa.createState());
						if(!st.containsKey(arc_h.getTo()))
							st.put(arc_h.getTo(), fa.createState());
						fa.addTransition(st.getValue(arc_h.getFrom()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");
					}
				}
			}
			if(!isSystem&&Options.backward){
				Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
				while(brel_it.hasNext()){
					Pair<FAState,FAState> p=brel_it.next();
					if((p.getLeft().getowner()==spec&&p.getRight().getowner()==spec)){
						if(!st.containsKey(p.getLeft().getID()))
							st.put(p.getLeft().getID(), fa.createState());
						if(!st.containsKey(p.getRight().getID()))
							st.put(p.getRight().getID(), fa.createState());
						fa.addTransition(st.getValue(p.getLeft().getID()), st.getValue(p.getRight().getID()), "0");
					}
				}
			}
			if(isSystem){
				Iterator<Pair<FAState,FAState>>frel_it=frel.iterator();
				while(frel_it.hasNext()){
					Pair<FAState,FAState> p=frel_it.next();
					if((p.getLeft().getowner()==system&&p.getRight().getowner()==system)){
						if(!st.containsKey(p.getLeft().getID()))
							st.put(p.getLeft().getID(), fa.createState());
						if(!st.containsKey(p.getRight().getID()))
							st.put(p.getRight().getID(), fa.createState());
						fa.addTransition(st.getValue(p.getRight().getID()), st.getValue(p.getLeft().getID()), "0");
					}
				}
			}
			SCC s=new SCC(fa);
			BitSet T=new BitSet();
			Iterator<FAState> s_it=s.getResult().iterator();
			while(s_it.hasNext()){
				T.set(st.getKey(s_it.next()));
			}
			Graph isolatedArcs=new Graph(Options.layered);
			isolatedArcs.addAll(h);			
			
			if(!isSystem&&Options.backward){
				Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
				while(brel_it.hasNext()){
					Pair<FAState,FAState> p=brel_it.next();
					if((p.getLeft().getowner()==spec&&p.getRight().getowner()==spec)){
						isolatedArcs.add(new Arc(p.getLeft().id,false,p.getRight().id));
					}
				}
			}
			if(isSystem){
				Iterator<Pair<FAState,FAState>>frel_it=frel.iterator();
				while(frel_it.hasNext()){
					Pair<FAState,FAState> p=frel_it.next();
					if((p.getLeft().getowner()==system&&p.getRight().getowner()==system)){
						isolatedArcs.add(new Arc(p.getRight().id,false,p.getLeft().id));
					}
				}
			}
			
			boolean changed=true;
			while(changed){
				changed=false;
				Graph isolatedArcsTemp=new Graph(Options.layered);
				if(!Options.layered){
					Iterator<Arc> arc_it=isolatedArcs.iterator();
					while(arc_it.hasNext()){
						Arc arc=arc_it.next();
						if(!T.get(arc.getTo())){
							isolatedArcsTemp.add(arc);
						}else{
							changed=true;
							T.set(arc.getFrom());
						}
					}
				}else{
					Iterator<Integer> left_it=isolatedArcs.leftStateIt();
					while(left_it.hasNext()){
						Iterator<Arc> arc_it=isolatedArcs.iterator(left_it.next());
						while(arc_it.hasNext()){
							Arc arc=arc_it.next();
							if(!T.get(arc.getTo())){
								isolatedArcsTemp.add(arc);
							}else{
								changed=true;
								T.set(arc.getFrom());
							}
						}
					}
				}
				isolatedArcs=isolatedArcsTemp;
			}
			if(isSystem)
				SysTail.put(h, T);
			else
				SpecTail.put(h, T);
		}
		if(isSystem)
			return SysTail.get(h);
		else
			return SpecTail.get(h);
	}	

	private boolean contains(Set<Pair<FAState,FAState>> rel, FAState s1, FAState s2){
		return rel.contains(new Pair<FAState,FAState>(s1,s2));
	}
	private boolean leftSubsume(Arc cur, Arc other){
		if(cur.getFrom()==other.getFrom()){
			if(!cur.getLabel()||other.getLabel()){
				if(cur.getTo()!=other.getTo()){
					if(frelM.get(new FAState(cur.getTo(),system)).contains(new FAState(other.getTo(), system ))){
						if(!frelM.get(new FAState(other.getTo(),system)).contains(new FAState(cur.getTo(), system ))){
							return true;
						}else if(cur.getTo()<other.getTo()){
							return true;
						}
					}
				}
			}
		}
		return false;
	}
	private boolean rightSubsume(Arc cur, Arc other, boolean l){
		if(l||brelM.get(new FAState(cur.getFrom(),spec)).contains(new FAState(other.getFrom(), spec ))){
			if(other.getFrom()>=cur.getFrom()||!brelM.get(new FAState(other.getFrom(),spec)).contains(new FAState(cur.getFrom(), spec ))){
				if(!cur.getLabel()||other.getLabel()){
					if(frelM.get(new FAState(cur.getTo(),spec)).contains(new FAState(other.getTo(), spec ))){
						if(other.getTo()>=cur.getTo()||!frelM.get(new FAState(other.getTo(),spec)).contains(new FAState(cur.getTo(), spec ))){
							return true;
						}
					}
				}
			}
		}
		return false;
	}
	
	private NewMetagraph min_plus(NewMetagraph NewMetagraph){
		debug("before min:"+NewMetagraph,verboseMin);
		Graph left_result=new Graph(Options.layered);
		if(!Options.layered){
			Iterator<Arc> arc_it1 =NewMetagraph.getLeft().iterator();
			while(arc_it1.hasNext()){
				Arc cur=arc_it1.next();
				boolean canAdd=true;
				Iterator<Arc> arc_it2 =NewMetagraph.getLeft().iterator();
				while(arc_it2.hasNext()){
					Arc other=arc_it2.next();
					if(leftSubsume(cur,other)){
						canAdd=false;
						break;
					}
				}			
				if(canAdd){
					left_result.add(cur);
				}
			}
			
		}else{
			Iterator<Integer> leftSt_it=NewMetagraph.getLeft().leftStateIt();
			while(leftSt_it.hasNext()){
				Iterator<Arc> arc_it1 =NewMetagraph.getLeft().iterator(leftSt_it.next());
				while(arc_it1.hasNext()){
					Arc cur=arc_it1.next();
					boolean canAdd=true;
					Iterator<Arc> arc_it2 =NewMetagraph.getLeft().iterator(cur.getFrom());
					while(arc_it2.hasNext()){
						Arc other=arc_it2.next();
						if(leftSubsume(cur,other)){
							canAdd=false;
							break;
						}
					}			
					if(canAdd){
						left_result.add(cur);
					}
				}
			}
		}

		Graph right_result=new Graph(Options.layered);
		if(!Options.layered){
			Iterator<Arc> arc_it1 =NewMetagraph.getRight().iterator();
			while(arc_it1.hasNext()){
				Arc cur=arc_it1.next();
				boolean canAdd=true;
				Iterator<Arc> arc_it2 =NewMetagraph.getRight().iterator();
				while(arc_it2.hasNext()){
					Arc other=arc_it2.next();
					if(cur.getFrom()==other.getFrom() && cur.getTo()==other.getTo())
						continue;
					if(rightSubsume(cur,other,false)){
						canAdd=false;
						break;						
					}
				}
				if(canAdd){
					right_result.add(cur);
				}
			}
		}else{
			Iterator<Integer> leftSt_it=NewMetagraph.getRight().leftStateIt();
			while(leftSt_it.hasNext()){
				Iterator<Arc> arc_it1 =NewMetagraph.getRight().iterator(leftSt_it.next());
				while(arc_it1.hasNext()){
					Arc cur=arc_it1.next();
					boolean canAdd=true;
					Iterator<FAState> bwLgSts_it= brelM.get(new FAState(cur.getFrom(),spec)).iterator();
					while(bwLgSts_it.hasNext()){
						FAState bwLgSt=bwLgSts_it.next();
						if(bwLgSt.getowner()!=spec)
							continue;
					
						Iterator<Arc> arc_it2 =NewMetagraph.getRight().iterator(bwLgSt.getID());
						while(arc_it2.hasNext()){
							Arc other=arc_it2.next();
							if(cur.getFrom()==other.getFrom() && cur.getTo()==other.getTo())
								continue;
							if(rightSubsume(cur,other,true)){
								canAdd=false;
								break;						
							}
						}
						if(!canAdd)
							break;
					}
					if(canAdd){
						right_result.add(cur);
					}
				}
			}
		}
		debug("after min:"+new Pair<Graph, Graph>(left_result,right_result),verboseMin);
		return new NewMetagraph(left_result,right_result,NewMetagraph.rpstring);
	}
	

	private ArrayList<NewMetagraph> buildSingleCharacterNewMetagraphs2(){
		TreeMap<Pair<Arc, Graph>, String> SGStringmap=new TreeMap<Pair<Arc, Graph>, String>(new NewSuperGraphComparator());
		ArrayList<Pair<Arc, Graph>> supergraphs=new ArrayList<Pair<Arc, Graph>>();
		Iterator<String> symbol_it=system.getAllTransitionSymbols().iterator();
		while(symbol_it.hasNext()){
			Graph graph=new Graph(Options.layered);
			String sym=symbol_it.next();
			Iterator<FAState> from_it=spec.states.iterator();
			while(from_it.hasNext()){
				FAState from=from_it.next();
				if(from.getNext(sym)!=null){
					Iterator<FAState> to_it=from.getNext(sym).iterator();
					while(to_it.hasNext()){
						FAState to=to_it.next();
						if(spec.F.contains(from)||spec.F.contains(to)){
							graph.add(new Arc(from.id,true,to.id));
						}else{
							graph.add(new Arc(from.id,false,to.id));
						}
					}
				}
			}

			from_it=system.states.iterator();
			while(from_it.hasNext()){
				FAState from=from_it.next();
				if(from.getNext(sym)!=null){
					Iterator<FAState> to_it=from.getNext(sym).iterator();
					while(to_it.hasNext()){
						FAState to=to_it.next();
						Arc left_arc;
						if(system.F.contains(from)||system.F.contains(to)){
							left_arc=new Arc(from.id,true,to.id);
						}else{
							left_arc=new Arc(from.id,false,to.id);
						}
						Pair<Arc, Graph> supergraph=new Pair<Arc, Graph>(left_arc,graph);
						ArrayList<Pair<Arc, Graph>> toRemove=new ArrayList<Pair<Arc, Graph>>();
						boolean canAdd=true;
						Iterator<Pair<Arc, Graph>> old_it=supergraphs.iterator();
						while(old_it.hasNext()){
							Pair<Arc, Graph> old=old_it.next();
							if(smallerThan(old, supergraph)){
								canAdd=false;
								break;
							}else if(smallerThan(supergraph, old)){
								toRemove.add(old);
							}
						}
						if(canAdd){
							if(Options.opt2)
									supergraph=smin_plus(supergraph);

							supergraphs.add(supergraph);
							supergraphs.removeAll(toRemove);
							if(Options.debug)
								SGStringmap.put(supergraph, sym);
						}
					}
				}
			}
		}
		TreeMap<Graph, NewMetagraph> MGmap=new TreeMap<Graph, NewMetagraph>();
		ArrayList<NewMetagraph> NewMetagraphs=new ArrayList<NewMetagraph>();
		Iterator<Pair<Arc, Graph>> SG_it=supergraphs.iterator();
		while(SG_it.hasNext()){
			Pair<Arc, Graph> SG=SG_it.next();
			if(MGmap.containsKey(SG.getRight())){
				MGmap.get(SG.getRight()).getLeft().add(SG.getLeft());
			}else{
				Graph left=new Graph(Options.layered);
				left.add(SG.getLeft());
				NewMetagraph MG=new NewMetagraph(left, SG.getRight());
				MGmap.put(MG.getRight(), MG);
				MG.rpstring=SGStringmap.get(SG);
				NewMetagraphs.add(MG);
			}
			
		}

		Iterator<NewMetagraph> Q1_it=NewMetagraphs.iterator();
		while(Q1_it.hasNext()){
			NewMetagraph g=Q1_it.next();
			if(!Options.layered){
				Iterator<Arc> left_it=g.getLeft().iterator();
				while(left_it.hasNext()){
					Arc left=left_it.next();
					boolean c1=C1(left,g.getRight());
					if(!c1){
						left.L=true;
						cntL++;
					}
				}
			}else{
				Iterator<Integer> leftSt_it=g.getLeft().leftStateIt();
				while(leftSt_it.hasNext()){
					Iterator<Arc> left_it=g.getLeft().iterator(leftSt_it.next());
					while(left_it.hasNext()){
						Arc left=left_it.next();
						boolean c1=C1(left,g.getRight());
						if(!c1){
							left.L=true;
							cntL++;
						}
					}
				}
			}
		}
		return NewMetagraphs;
	}

	private ArrayList<NewMetagraph> buildSingleCharacterNewMetagraphs(){
		ArrayList<NewMetagraph> NewMetagraphs=new ArrayList<NewMetagraph>();
		Iterator<String> symbol_it=system.getAllTransitionSymbols().iterator();
		while(symbol_it.hasNext()){
			Graph rightGraph=new Graph(Options.layered);
			String sym=symbol_it.next();
			Iterator<FAState> from_it=spec.states.iterator();
			while(from_it.hasNext()){
				FAState from=from_it.next();
				if(from.getNext(sym)!=null){
					Iterator<FAState> to_it=from.getNext(sym).iterator();
					while(to_it.hasNext()){
						FAState to=to_it.next();
						if(spec.F.contains(from)||spec.F.contains(to)){
							rightGraph.add(new Arc(from.id,true,to.id));
						}else{
							rightGraph.add(new Arc(from.id,false,to.id));
						}
					}
				}
			}

			Graph leftGraph=new Graph(Options.layered);
			from_it=system.states.iterator();
			while(from_it.hasNext()){
				FAState from=from_it.next();
				if(from.getNext(sym)!=null){
					Iterator<FAState> to_it=from.getNext(sym).iterator();
					while(to_it.hasNext()){
						FAState to=to_it.next();
						if(system.F.contains(from)||system.F.contains(to)){
							leftGraph.add(new Arc(from.id,true,to.id));
						}else{
							leftGraph.add(new Arc(from.id,false,to.id));
						}
					}
				}
			}
			NewMetagraph f=min_plus(new NewMetagraph(leftGraph,rightGraph));
			f.rpstring=sym;

			ArrayList<NewMetagraph> toRemove=new ArrayList<NewMetagraph>();
			ArrayList<NewMetagraph> toAdd=new ArrayList<NewMetagraph>();

			Iterator<NewMetagraph> NewMetagraph_it=NewMetagraphs.iterator();
			while(NewMetagraph_it.hasNext()){
				NewMetagraph p=NewMetagraph_it.next();
				if(Options.layered)
					cleanLayered(f, p, (short) 1);
				else
					clean(f, p, (short) 1);
				if(f.getLeft().size()==0)
					break;

				Graph left=new Graph(Options.layered);
				left.addAll(p.getLeft());
				Graph right=new Graph(Options.layered);
				right.addAll(p.getRight());
				NewMetagraph p_new=new NewMetagraph(left, right,p.rpstring);
				if(Options.layered)
					cleanLayered(p_new, f, (short) 3);
				else
					clean(p_new, f, (short) 3);
				if(p_new.compareTo(p)!=0){
					toRemove.add(p);
					if(p_new.getLeft().size()>0)
						toAdd.add(p_new);
				}
			}			
			if(f.getLeft().size()!=0){
				NewMetagraphs.add(f);
				NewMetagraphs.removeAll(toRemove);
				NewMetagraphs.addAll(toAdd);
			}
		}


		
		Iterator<NewMetagraph> Q1_it=NewMetagraphs.iterator();
		while(Q1_it.hasNext()){
			NewMetagraph g=Q1_it.next();
			if(!Options.layered){
				Iterator<Arc> left_it=g.getLeft().iterator();
				while(left_it.hasNext()){
					Arc left=left_it.next();
					boolean c1=C1(left,g.getRight());
					if(!c1){
						left.L=true;
						cntL++;
					}
				}
			}else{
				Iterator<Integer> leftSt_it=g.getLeft().leftStateIt();
				while(leftSt_it.hasNext()){
					Iterator<Arc> left_it=g.getLeft().iterator(leftSt_it.next());
					while(left_it.hasNext()){
						Arc left=left_it.next();
						boolean c1=C1(left,g.getRight());
						if(!c1){
							left.L=true;
							cntL++;
						}
					}
				}
			}
		}
		return NewMetagraphs;
	}
	
	private Pair<Arc, Graph> smin_plus(Pair<Arc, Graph> supergraph){
		debug("before min:"+supergraph,verboseMin);
		Graph result=new Graph(Options.layered);
		if(!Options.layered){
			Iterator<Arc> arc_it1 =supergraph.getRight().iterator();
			while(arc_it1.hasNext()){
				Arc cur=arc_it1.next();
				boolean canAdd=true;
				Iterator<Arc> arc_it2 =supergraph.getRight().iterator();
				while(arc_it2.hasNext()){
					Arc other=arc_it2.next();
					if(cur.getFrom()==other.getFrom() && cur.getTo()==other.getTo())//"cur" and "other" refer to the same arc
						continue;
					if(brelM.get(new FAState(cur.getFrom(),spec)).contains(new FAState(other.getFrom(), spec ))){
						if(other.getFrom()>=cur.getFrom()||!brelM.get(new FAState(other.getFrom(),spec)).contains(new FAState(cur.getFrom(), spec ))){
						   //if the id of cur.from is larger, then cur.from and other.from cannot be bw-sim equivalent
						   if(!cur.getLabel()||other.getLabel()){
							   if(frelM.get(new FAState(cur.getTo(),spec)).contains(new FAState(other.getTo(), spec ))){
								   if(other.getTo()>=cur.getTo()||!contains(frel, new FAState(other.getTo(),spec),new FAState(cur.getTo(), spec ))){
									    //if the id of cur.to is larger, then cur.to and other.to cannot be bw-sim equivalent
										canAdd=false;
										break;
								   }
							   }
						   }
						}
					}
				}
				if(canAdd){
					result.add(cur);
				}
			}
		}else{//layered version
			Iterator<Integer> leftSt_it=supergraph.getRight().leftStateIt();
			while(leftSt_it.hasNext()){
				int leftSt=leftSt_it.next();
				Iterator<Arc> arc_it1 =supergraph.getRight().iterator(leftSt);
				while(arc_it1.hasNext()){
					Arc cur=arc_it1.next();
					boolean canAdd=true;
					Iterator<FAState> bwLgSts_it= brelM.get(new FAState(cur.getFrom(),spec)).iterator();
					while(bwLgSts_it.hasNext()){
						FAState bwLgSt=bwLgSts_it.next();
						if(bwLgSt.getowner()!=spec)
							continue;

						Iterator<Arc> arc_it2 =supergraph.getRight().iterator(bwLgSt.getID());
						while(arc_it2.hasNext()){
							Arc other=arc_it2.next();
							if(cur.getFrom()==other.getFrom()&&cur.getTo()==other.getTo())
								continue;
							if(other.getFrom()>=cur.getFrom()||!brelM.get(new FAState(other.getFrom(),spec)).contains(new FAState(cur.getFrom(), spec ))){
							   //if the id of cur.from is larger, then cur.from and other.from cannot be bw-sim equivalent
							   if(!cur.getLabel()||other.getLabel()){
								   if(frelM.get(new FAState(cur.getTo(),spec)).contains(new FAState(other.getTo(), spec ))){
									   if(other.getTo()>=cur.getTo()||!contains(frel, new FAState(other.getTo(),spec),new FAState(cur.getTo(), spec ))){
										    //if the id of cur.to is larger, then cur.to and other.to cannot be bw-sim equivalent
											canAdd=false;
											break;
									   }
								   }
							   }
							}
						}
					}
					if(canAdd){
						result.add(cur);
					}
				}
			}
		}
		debug("after min:"+new Pair<Arc, Graph>(supergraph.getLeft(),result),verboseMin);
		return new Pair<Arc, Graph>(supergraph.getLeft(),result);
	}
	
	private void labelArc(boolean C1, boolean LLabeled, Arc arc, Graph right){
		arc.R=true;
		if(C1){
			if(LLabeled){
				if(!C1(arc, right)){
					arc.L=true;
					debug("Add L label to "+arc,verboseC1);
					cntL++;
				}
			}
		}else{
			arc.L=true;
			debug("Add L label to "+arc,verboseC1);
			cntL++;
		}
	}
	private NewMetagraph compose(NewMetagraph g, NewMetagraph h, Graph oriL){
		Graph right=new Graph(Options.layered);
		if(!Options.layered){
			Iterator<Arc> arc_g_it=g.getRight().iterator();
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				Iterator<Arc> arc_h_it=h.getRight().iterator();
				while(arc_h_it.hasNext()){
					Arc arc_h=arc_h_it.next();
					boolean connected=false;
					if(Options.backward &&contains(brel, new FAState(arc_g.getTo(),spec),new FAState(arc_h.getFrom(), spec )))
						connected=true;
					else if(!Options.backward&&arc_g.getTo()==arc_h.getFrom())
						connected=true;
					if(connected){
						if(arc_g.getLabel()||arc_h.getLabel()){
							right.add(new Arc(arc_g.getFrom(),true,arc_h.getTo()));
							right.remove(new Arc(arc_g.getFrom(),false,arc_h.getTo()));
						}else{
							if(!right.contains(new Arc(arc_g.getFrom(),true,arc_h.getTo()))){
								right.add(new Arc(arc_g.getFrom(),false,arc_h.getTo()));
							}
						}
					}
				}
			}
		}else{
			Iterator<Integer> leftSt_it=g.getRight().leftStateIt();
			while(leftSt_it.hasNext()){
				Iterator<Arc> arc_g_it=g.getRight().iterator(leftSt_it.next());
				while(arc_g_it.hasNext()){
					Arc arc_g=arc_g_it.next();
					Iterator<FAState> bwLgSts_it=brelM.get(new FAState(arc_g.getTo(),spec)).iterator();
					while(bwLgSts_it.hasNext()){
						FAState bwLgSt=bwLgSts_it.next();
						if(bwLgSt.getowner()!=spec)
							continue;
						Iterator<Arc> arc_h_it=h.getRight().iterator(bwLgSt.getID());
						while(arc_h_it.hasNext()){
							Arc arc_h=arc_h_it.next();
							if(arc_g.getLabel()||arc_h.getLabel()){
								right.add(new Arc(arc_g.getFrom(),true,arc_h.getTo()));
								right.remove(new Arc(arc_g.getFrom(),false,arc_h.getTo()));
							}else{
								if(!right.contains(new Arc(arc_g.getFrom(),true,arc_h.getTo()))){
									right.add(new Arc(arc_g.getFrom(),false,arc_h.getTo()));
								}
							}
						}
					}
					
					
				}				
			}
		}
		Graph left=new Graph(Options.layered);
		if(!Options.layered){
			Iterator<Arc> arc_i_it=g.getLeft().iterator();
			while(arc_i_it.hasNext()){
				Arc arc_i=arc_i_it.next();
				boolean LLabeled=oriL.contains(arc_i);
				Iterator<Arc> arc_h_it=h.getLeft().iterator();
				while(arc_h_it.hasNext()){
					Arc arc_h=arc_h_it.next();
					if(arc_i.getTo()==arc_h.getFrom()){
						if(arc_i.getLabel()||arc_h.getLabel()){
							Arc newarc=new Arc(arc_i.getFrom(),true,arc_h.getTo());
							left.add(newarc);
							left.remove(new Arc(arc_i.getFrom(),false,arc_h.getTo()));
							labelArc(Options.C1,LLabeled,newarc,right);
						}else{
							Arc newarc=new Arc(arc_i.getFrom(),false,arc_h.getTo());
							if(!left.contains(new Arc(arc_i.getFrom(),true,arc_h.getTo()))){
								left.add(newarc);
								labelArc(Options.C1,LLabeled,newarc,right);
							}
						}
					}				
				}			
			}
		}else{
			Iterator<Integer> leftSt_it=g.getLeft().leftStateIt();
			while(leftSt_it.hasNext()){
				int leftSt=leftSt_it.next();
				Iterator<Arc> arc_i_it=g.getLeft().iterator(leftSt);
				while(arc_i_it.hasNext()){
					Arc arc_i=arc_i_it.next();
					boolean LLabeled=oriL.contains(arc_i);
					Iterator<Arc> arc_h_it=h.getLeft().iterator(arc_i.getTo());
					if(arc_h_it!=null){
						while(arc_h_it.hasNext()){
							Arc arc_h=arc_h_it.next();
							if(arc_i.getLabel()||arc_h.getLabel()){
								Arc newarc=new Arc(arc_i.getFrom(),true,arc_h.getTo());
								left.add(newarc);
								left.remove(new Arc(arc_i.getFrom(),false,arc_h.getTo()));
								labelArc(Options.C1,LLabeled,newarc,right);
							}else{
								Arc newarc=new Arc(arc_i.getFrom(),false,arc_h.getTo());
								if(!left.contains(new Arc(arc_i.getFrom(),true,arc_h.getTo()))){
									left.add(newarc);
									labelArc(Options.C1,LLabeled,newarc,right);
								}
							}
						}			
					}
				}
			}
		}
		if(left.size()==0){
			return null;
		}else{
			return new NewMetagraph(left,right,g.rpstring+h.rpstring);
		}
	}

	boolean clean(NewMetagraph src, NewMetagraph tgt, short type){
		
		Iterator<Arc> arc_g_it=tgt.getRight().iterator();
		while(arc_g_it.hasNext()){
			Arc arc_g=arc_g_it.next();
			boolean has_larger=false;
			Iterator<Arc> arc_h_it=src.getRight().iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				boolean fromOkay=false;
				if(Options.backward &&brel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getFrom(),spec),new FAState(arc_h.getFrom(), spec ))))
					fromOkay=true;
				else if(!Options.backward&&arc_g.getFrom()==arc_h.getFrom())
					fromOkay=true;
				if(fromOkay){
					if(!arc_g.getLabel()||arc_h.getLabel()){
						if(frelM.get(new FAState(arc_g.getTo(),spec)).contains(new FAState(arc_h.getTo(),spec))){
							has_larger=true;
							break;
						}
					}
				}
			}			
			if(!has_larger){
				return true;
			}
		}
		debug("Before clean"+type+" :"+src+" (w.r.t. "+tgt+")",verboseC1);
		//tgt.right smaller than src.right

		if(type!=3 && src.getRight().size()==tgt.getRight().size()){
			arc_g_it=src.getRight().iterator();
			boolean has_larger=true;
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				has_larger=false;
				Iterator<Arc> arc_h_it=tgt.getRight().iterator();
				while(arc_h_it.hasNext()){
					Arc arc_h=arc_h_it.next();
					boolean fromOkay=false;
					if(Options.backward &&brel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getFrom(),spec),new FAState(arc_h.getFrom(), spec ))))
						fromOkay=true;
					else if(!Options.backward&&arc_g.getFrom()==arc_h.getFrom())
						fromOkay=true;
					if(fromOkay){
						if(!arc_g.getLabel()||arc_h.getLabel()){
							if(frelM.get(new FAState(arc_g.getTo(),spec)).contains(new FAState(arc_h.getTo(),spec))){
								has_larger=true;
								break;
							}
						}
					}
				}			
				if(!has_larger){
					break;
				}
			}
			if(has_larger){
				//src.right := tgt.right
				src.getRight().clear();
				src.getRight().addAll(tgt.getRight());
			}
		}

		boolean regain=false;
		Graph left=new Graph(Options.layered);
		arc_g_it=src.getLeft().iterator();
		while(arc_g_it.hasNext()){
			Arc arc_g=arc_g_it.next();
			boolean has_larger=false;
			Iterator<Arc> arc_h_it=tgt.getLeft().iterator();
			Arc arc_h = null;
			while(arc_h_it.hasNext()){
				arc_h=arc_h_it.next();
				if(arc_g.getFrom()==arc_h.getFrom()){
					if(!arc_g.getLabel()||arc_h.getLabel()){
						if(frelM.get(new FAState(arc_g.getTo(),system)).contains(new FAState(arc_h.getTo(),system))){
							has_larger=true;
							break;
						}
					}
				}
			}			
			if(!has_larger){
				left.add(arc_g);
			}else if(type==2){
				if(arc_g.L){
					arc_h.L=true;
					cntL++;
					regain=true;
				}
			}
		}
		
		
		
		
//		if(src.toString().compareTo(tgt.toString())==0&& left.size()!=0)
//			System.out.println();
		src.getLeft().clear();
		src.getLeft().addAll(left);
		debug("After clean"+type+" :"+src,verboseC1);
		if(regain)
			debug("Regain L :"+tgt,verboseC1);
		return !regain;
	
	}
	
	//clean NewMetagraph src w.r.t tgt
	//if tgt.right not smaller than src.right then does nothing
	boolean cleanLayered(NewMetagraph src, NewMetagraph tgt, short type){
		Iterator<Integer> leftSt_it=tgt.getRight().leftStateIt();
		while(leftSt_it.hasNext()){
			Iterator<Arc> arc_g_it=tgt.getRight().iterator(leftSt_it.next());
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				boolean has_larger=false;
				Iterator<FAState> bwLgSts_it=brelM.get(new FAState(arc_g.getFrom(),spec)).iterator();
				while(bwLgSts_it.hasNext()){
					FAState bwLgSt=bwLgSts_it.next();
					if(bwLgSt.getowner()!=spec)
						continue;
					Iterator<Arc> arc_h_it=src.getRight().iterator(bwLgSt.getID());
					
					while(arc_h_it.hasNext()){
						Arc arc_h=arc_h_it.next();
						if(!arc_g.getLabel()||arc_h.getLabel()){
						    // if ((frelM.get(new FAState(arc_g.getTo(),spec)).contains(new FAState(arc_h.getTo(),spec))) != myfsim[arc_g.getTo()][arc_h.getTo()]) System.out.println("ERROR");
							if(myfsim[arc_g.getTo()][arc_h.getTo()]
							   //frelM.get(new FAState(arc_g.getTo(),spec)).contains(new FAState(arc_h.getTo(),spec))
){
								has_larger=true;
								break;
							}
						}
					}
					if(has_larger)
						break;
				}
				
				if(!has_larger){
					return true;
				}
			}
		}

		debug("Before clean"+type+" :"+src+" (w.r.t. "+tgt+")",100);
		//tgt.right smaller than src.right
		if(type!=3 && src.getRight().size()==tgt.getRight().size()){
			boolean has_larger=false;
			leftSt_it=src.getRight().leftStateIt();
			while(leftSt_it.hasNext()){
				int leftSt=leftSt_it.next();
				Iterator<Arc> arc_g_it=src.getRight().iterator(leftSt);
				while(arc_g_it.hasNext()){
					Arc arc_g=arc_g_it.next();
					has_larger=false;
					Iterator<FAState> bwLgSts_it=brelM.get(new FAState(arc_g.getFrom(),spec)).iterator();
					while(bwLgSts_it.hasNext()){
						FAState bwLgSt=bwLgSts_it.next();
						if(bwLgSt.getowner()!=spec)
							continue;
						Iterator<Arc> arc_h_it=tgt.getRight().iterator(bwLgSt.getID());
						while(arc_h_it.hasNext()){
							Arc arc_h=arc_h_it.next();
							if(!arc_g.getLabel()||arc_h.getLabel()){
								if(myfsim[arc_g.getTo()][arc_h.getTo()]
								   //frelM.get(new FAState(arc_g.getTo(),spec)).contains(new FAState(arc_h.getTo(),spec))
){
									has_larger=true;
									break;
								}
							}
						}
						if(has_larger)
							break;
					}
					if(!has_larger)
						break;
				}
				if(!has_larger)
					break;
			}
			if(has_larger){
				//src.right := tgt.right
				src.getRight().clear();
				src.getRight().addAll(tgt.getRight());
			}
		}
		boolean regain=false;
		Graph left=new Graph(Options.layered);

		leftSt_it=src.getLeft().leftStateIt();
		while(leftSt_it.hasNext()){
			Iterator<Arc> arc_g_it=src.getLeft().iterator(leftSt_it.next());
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				boolean has_larger=false;
				Iterator<Arc> arc_h_it=tgt.getLeft().iterator(arc_g.getFrom());
				Arc arc_h = null;
				while(arc_h_it.hasNext()){
					arc_h=arc_h_it.next();
					if(!arc_g.getLabel()||arc_h.getLabel()){
					    //if ((frelM.get(new FAState(arc_g.getTo(),system)).contains(new FAState(arc_h.getTo(),system))) != myfsim2[arc_g.getTo()][arc_h.getTo()]) System.out.println("ERROR");
						if(myfsim2[arc_g.getTo()][arc_h.getTo()]
						   //frelM.get(new FAState(arc_g.getTo(),system)).contains(new FAState(arc_h.getTo(),system))
){
							has_larger=true;
							break;
						}
					}
				}			
				if(!has_larger){
					left.add(arc_g);
				}else if(type==2){
					if(arc_g.L){
						arc_h.L=true;
						cntL++;
						regain=true;
					}
				}
			}
		}
		src.getLeft().clear();
		src.getLeft().addAll(left);
		debug("After clean"+type+" :"+src,100);
		if(regain)
			debug("Regain L :"+tgt,100);
		return !regain;
	
	}
	
	boolean smallerThan(Pair<Arc, Graph> old, Pair<Arc, Graph> old2){
		Arc old_arc=old.getLeft();
		Arc old2_arc=old2.getLeft();
		if(!(old_arc.getFrom()==old2_arc.getFrom() && (old_arc.getLabel() ||!old2_arc.getLabel()) &&
				frelM.get(new FAState(old2_arc.getTo(),system)).contains(new FAState(old_arc.getTo(),system)))){
			return false;
		}
		if(!Options.layered){
			Iterator<Arc> arc_g_it=old.getRight().iterator();
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				boolean has_larger=false;
				Iterator<Arc> arc_h_it=old2.getRight().iterator();
				while(arc_h_it.hasNext()){
					Arc arc_h=arc_h_it.next();
					boolean fromOkay=false;
					if(Options.backward &&brel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getFrom(),spec),new FAState(arc_h.getFrom(), spec ))))
						fromOkay=true;
					else if(!Options.backward&&arc_g.getFrom()==arc_h.getFrom())
						fromOkay=true;
					if(fromOkay){
						if(!arc_g.getLabel()||arc_h.getLabel()){
							if(frelM.get(new FAState(arc_g.getTo(),spec)).contains(new FAState(arc_h.getTo(),spec))){
								has_larger=true;
								break;
							}
						}
					}
				}			
				if(!has_larger){
					return false;
				}
			}
		}else{
			Iterator<Integer> leftSt_it=old.getRight().leftStateIt();
			while(leftSt_it.hasNext()){
				int leftSt=leftSt_it.next();
				Iterator<Arc> arc_g_it =old.getRight().iterator(leftSt);
				while(arc_g_it.hasNext()){
					Arc arc_g=arc_g_it.next();
					boolean has_larger=false;
					Iterator<FAState> bwLgSts_it= brelM.get(new FAState(arc_g.getFrom(),spec)).iterator();
					while(bwLgSts_it.hasNext()){
						FAState bwLgSt=bwLgSts_it.next();
						if(bwLgSt.getowner()!=spec)
							continue;
						Iterator<Arc> arc_h_it=old2.getRight().iterator(bwLgSt.getID());
						while(arc_h_it.hasNext()){
							Arc arc_h=arc_h_it.next();
							if(!arc_g.getLabel()||arc_h.getLabel()){
								if(frelM.get(new FAState(arc_g.getTo(),spec)).contains(new FAState(arc_h.getTo(),spec))){
									has_larger=true;
									break;
								}
							}
						}
					}
					if(!has_larger){
						return false;
					}
				}
			}
		}
		return true;
	}

    public boolean[][] myfsim;
    public boolean[][] myfsim2;

	public void inclusionTest(){
		if(Options.rd){
			spec=removeDead(spec);
			system=removeDead(system);
		}
		if(Options.rd){
			System.out.println("Aut A (dead states removed): # of Trans. "+system.trans+", # of States "+system.states.size()+".");
			System.out.println("Aut B (dead states removed): # of Trans. "+spec.trans+", # of States "+spec.states.size()+".");
		}
		
		if(Options.qr){
			int qcount=0;
		    while(qcount < 10 && computeSim2()) { 
				if(Options.C1&&frel.contains(new Pair<FAState,FAState>(system.getInitialState(),spec.getInitialState()))){
//					System.out.print(system.trans+",\t "+system.states.size()+",\t ");
//					System.out.print(spec.trans+",\t "+spec.states.size()+",\t ");
//					System.out.print("F,\t ");
					return;
				}
				++qcount;
		    }
		}

		this.computeSim();

		// Now get matrix for frel.
		int mydim=spec.states.size();
		myfsim = new boolean[mydim][mydim];
		for (int i=0; i<mydim; i++)
		    for (int j=0; j<mydim; j++) myfsim[i][j]=false;
		int mydim2=system.states.size();
		myfsim2 = new boolean[mydim2][mydim2];
		for (int i=0; i<mydim2; i++)
		    for (int j=0; j<mydim2; j++) myfsim2[i][j]=false;

		Iterator<Pair<FAState, FAState>> frel_it=frel.iterator();
		while(frel_it.hasNext()){
			Pair<FAState, FAState> rel=frel_it.next();
			if (rel.getLeft().getowner()==spec && rel.getRight().getowner()==spec)
			    myfsim[rel.getLeft().getID()][rel.getRight().getID()]=true;
			if (rel.getLeft().getowner()==system && rel.getRight().getowner()==system)
			    myfsim2[rel.getLeft().getID()][rel.getRight().getID()]=true;
			}
		
//		System.out.print(system.trans+",\t "+system.states.size()+",\t ");
//		System.out.print(spec.trans+",\t "+spec.states.size()+",\t ");
		if(frel.contains(new Pair<FAState,FAState>(system.getInitialState(),spec.getInitialState()))){
//			System.out.print("F,\t ");
			if(Options.C1)
				return;
		}else{
//			System.out.print("N,\t ");
		}

		debug("Aut A:",100);
		debug(system.toString(),100);
		debug("Aut B:",100);
		debug(spec.toString(),100);
		debug("FSim="+showSim(frel),100);
		if(Options.backward)
		debug("BSim="+showSim(brel),100);
		this.included=included();
	}
	public long getCpuTime( ) {
	    ThreadMXBean bean = ManagementFactory.getThreadMXBean( );
	    return bean.isCurrentThreadCpuTimeSupported( ) ?
	        bean.getCurrentThreadCpuTime( ) : 0L;
	}
	
	public boolean isIncluded(){
		return included;
	}
	public void run(){
		runTime=getCpuTime();
			inclusionTest();
		runTime=getCpuTime()-runTime;
	}
	
	public long getRunTime(){
		return runTime;
	}

	boolean computeSim2(){
	    int oldspec=spec.states.size();
	    int oldsys=system.states.size();

		ArrayList<FAState> all_states=new ArrayList<FAState>();
		HashSet<String> alphabet=new HashSet<String>();
		
		all_states.addAll(spec.states);
		alphabet.addAll(spec.alphabet);

		all_states.addAll(system.states);
		alphabet.addAll(system.alphabet);
		
		
	    boolean result=false;
		FAState[] states=all_states.toArray(new FAState[0]);		
		boolean[] isFinal = new boolean[states.length];
		boolean[] isInit = new boolean[states.length];
		boolean[][] fsim = new boolean[states.length][states.length];
		boolean[][] bsim = new boolean[states.length][states.length];
		for(int i=0;i<states.length;i++){			
			isFinal[i] = states[i].getowner().F.contains(states[i]);
			isInit[i] =states[i].getowner().getInitialState().compareTo(states[i])==0;
		}
		for(int i=0;i<states.length;i++){
			for(int j=i;j<states.length;j++){
				fsim[i][j] = (!isFinal[i] || isFinal[j]) && states[j].fw_covers(states[i]);
				fsim[j][i] = (isFinal[i] || !isFinal[j]) && states[i].fw_covers(states[j]);
				if(Options.backward){
					bsim[i][j] = (!isInit[i] || isInit[j]) && (!isFinal[i] || isFinal[j]) && states[j].bw_covers(states[i]);
					bsim[j][i] = (isInit[i] || !isInit[j]) && (isFinal[i] || !isFinal[j]) && states[i].bw_covers(states[j]);
				}
			}
		}
		Simulation sim = new Simulation();
		frel = sim.FastFSimRelNBW(spec,system,fsim);
		if(Options.C1&&frel.contains(new Pair<FAState,FAState>(system.getInitialState(),spec.getInitialState()))){
			return true;
		}		

		if(Options.fplus){
			Set<Pair<FAState,FAState>> temp=new TreeSet<Pair<FAState,FAState>>(new StatePairComparator());
			for(int i=0;i<states.length;i++){
				for(int j=0;j<states.length;j++){
					if(isFinal[i]&&!isFinal[j]){
						Iterator<String> alphabet_it=states[i].nextIt();
						boolean covered=true;
						while(alphabet_it.hasNext()){
							String a=alphabet_it.next();
							if(states[i].getNext(a)!=null && states[j].getNext(a)==null){
								covered=false;
								break;
							}
								
							Iterator<FAState> small_it=states[i].getNext(a).iterator();
							while(small_it.hasNext()){
								FAState small=small_it.next();
								covered=false;
								if(states[j].getNext(a)!=null){
									Iterator<FAState> big_it=states[j].getNext(a).iterator();
									while(big_it.hasNext()){
										FAState big=big_it.next();
										if(frel.contains(new Pair<FAState, FAState>(small, big))&& big.getowner().F.contains(big)){
											covered=true;
											break;
										}
									}
								}
								if(!covered)
									break;
							}
							if(!covered)
								break;
						}
						if(covered){
							temp.add(new Pair<FAState,FAState>(states[i],states[j]));
						}
					}
				}
			}
			frel.addAll(temp);
		}
		spec=quotient(spec, frel);
		system=quotient(system, frel);
		//frelM has to be recomputed
		if(Options.backward){
			all_states.clear();
			all_states.addAll(spec.states);
			all_states.addAll(system.states);
			states=all_states.toArray(new FAState[0]);		
			isFinal = new boolean[states.length];
			isInit = new boolean[states.length];
			bsim = new boolean[states.length][states.length];
			for(int i=0;i<states.length;i++){			
				isFinal[i] = states[i].getowner().F.contains(states[i]);
				isInit[i] =states[i].getowner().getInitialState().compareTo(states[i])==0;
			}
			for(int i=0;i<states.length;i++){
				for(int j=i;j<states.length;j++){
					bsim[i][j] = (!isInit[i] || isInit[j]) && (!isFinal[i] || isFinal[j]) && states[j].bw_covers(states[i]);
					bsim[j][i] = (isInit[i] || !isInit[j]) && (isFinal[i] || !isFinal[j]) && states[i].bw_covers(states[j]);
				}
			}
			brel = sim.FastBSimRelNBW(spec,system,bsim);
		    spec=quotient(spec, brel);
		    system=quotient(system, brel);
		    //brelM has to be recomputed
		}
	    if (oldspec > spec.states.size() || oldsys > system.states.size()) 
	    	result=true;
	    return(result);
	}	

	void computeSim(){

		ArrayList<FAState> all_states=new ArrayList<FAState>();
		HashSet<String> alphabet=new HashSet<String>();
		
		all_states.addAll(spec.states);
		alphabet.addAll(spec.alphabet);

		all_states.addAll(system.states);
		alphabet.addAll(system.alphabet);
		
		
		FAState[] states=all_states.toArray(new FAState[0]);		
		boolean[] isFinal = new boolean[states.length];
		boolean[] isInit = new boolean[states.length];
		boolean[][] fsim = new boolean[states.length][states.length];
		boolean[][] bsim = new boolean[states.length][states.length];
		// sim[u][v]=true iff v in sim(u) iff v simulates u
		for(int i=0;i<states.length;i++){			
			isFinal[i] = states[i].getowner().F.contains(states[i]);
			isInit[i] =states[i].getowner().getInitialState().compareTo(states[i])==0;
		}
		for(int i=0;i<states.length;i++){
			for(int j=i;j<states.length;j++){
				fsim[i][j] = (!isFinal[i] || isFinal[j]) && states[j].fw_covers(states[i]);
				fsim[j][i] = (isFinal[i] || !isFinal[j]) && states[i].fw_covers(states[j]);
			}
		}
		Simulation sim = new Simulation();
		frel = sim.FastFSimRelNBW(spec,system,fsim);
		if(Options.C1){
			if(frel.contains(new Pair<FAState,FAState>(system.getInitialState(),spec.getInitialState()))){
				return;
			}
		}

		if(Options.fplus){
			Set<Pair<FAState,FAState>> temp=new TreeSet<Pair<FAState,FAState>>(new StatePairComparator());
			for(int i=0;i<states.length;i++){
				for(int j=0;j<states.length;j++){
					if(isFinal[i]&&!isFinal[j]){
						Iterator<String> alphabet_it=states[i].nextIt();
						boolean covered=true;
						while(alphabet_it.hasNext()){
							String a=alphabet_it.next();
							if(states[i].getNext(a)!=null && states[j].getNext(a)==null){
								covered=false;
								break;
							}
								
							Iterator<FAState> small_it=states[i].getNext(a).iterator();
							while(small_it.hasNext()){
								FAState small=small_it.next();
								covered=false;
								if(states[j].getNext(a)!=null){
									Iterator<FAState> big_it=states[j].getNext(a).iterator();
									while(big_it.hasNext()){
										FAState big=big_it.next();

										if(frel.contains(new Pair<FAState, FAState>(small, big))&& big.getowner().F.contains(big)){
											covered=true;
											break;
										}
									}
								}
								if(!covered)
									break;
							}
							if(!covered)
								break;
						}
						if(covered){
							temp.add(new Pair<FAState,FAState>(states[i],states[j]));
						}
					}
				}
			}
			frel.addAll(temp);
		}
		
		if(Options.quotient){
			spec=quotient(spec, frel);
			system=quotient(system, frel);
		}

		frelM=new TreeMap<FAState, Set<FAState>>();
		Iterator<Pair<FAState, FAState>> frel_it=frel.iterator();
		while(frel_it.hasNext()){
			Pair<FAState, FAState> rel=frel_it.next();
			if(!frelM.containsKey(rel.getLeft())){
				Set<FAState> sts=new TreeSet<FAState>();
				frelM.put(rel.getLeft(), sts);
			}
			frelM.get(rel.getLeft()).add(rel.getRight());
		}
		
		
		brelM=new TreeMap<FAState, Set<FAState>>();
		all_states.clear();
		all_states.addAll(spec.states);
		all_states.addAll(system.states);
		states=all_states.toArray(new FAState[0]);		
		
		if(Options.backward){
			isFinal = new boolean[states.length];
			isInit = new boolean[states.length];
			bsim = new boolean[states.length][states.length];
			for(int i=0;i<states.length;i++){			
				isFinal[i] = states[i].getowner().F.contains(states[i]);
				isInit[i] =states[i].getowner().getInitialState().compareTo(states[i])==0;
			}
			for(int i=0;i<states.length;i++){
				for(int j=i;j<states.length;j++){
					bsim[i][j] = (!isInit[i] || isInit[j]) && (!isFinal[i] || isFinal[j]) && states[j].bw_covers(states[i]);
					bsim[j][i] = (isInit[i] || !isInit[j]) && (isFinal[i] || !isFinal[j]) && states[i].bw_covers(states[j]);
				}
			}
			brel = sim.FastBSimRelNBW(spec,system,bsim);
			Iterator<Pair<FAState, FAState>> brel_it=brel.iterator();
			while(brel_it.hasNext()){
				Pair<FAState, FAState> rel=brel_it.next();
				if(!brelM.containsKey(rel.getLeft())){
					Set<FAState> sts=new TreeSet<FAState>();
					brelM.put(rel.getLeft(), sts);
				}
				brelM.get(rel.getLeft()).add(rel.getRight());
			}
		}else{
			brel = new TreeSet<Pair<FAState,FAState>>(new StatePairComparator());
			for(int i=0;i<states.length;i++){
				Set<FAState> sts=new TreeSet<FAState>();
				sts.add(states[i]);
				brelM.put(states[i], sts);
				brel.add(new Pair<FAState, FAState>(states[i],states[i]));
			}
		}
	}	
	
	private FiniteAutomaton removeDead(FiniteAutomaton aut) {
		FiniteAutomaton result=new FiniteAutomaton();
		result.name=aut.name;
		HashSet<FAState> finalReachable=new HashSet<FAState>();

		SCC s=new SCC(aut, true);
		finalReachable.addAll(s.getResult());
		
		while(true){
			HashSet<FAState> toAdd=new HashSet<FAState>();
			Iterator<FAState> finalReachable_it=finalReachable.iterator();
			while(finalReachable_it.hasNext()){
				FAState end=finalReachable_it.next();
				toAdd.addAll(end.getPre());
			}
			toAdd.removeAll(finalReachable);
			if(toAdd.size()==0)
				break;
			finalReachable.addAll(toAdd);
		}

		HashMap<FAState, FAState> stMap=new HashMap<FAState, FAState>();
		HashSet<FAState> toProcess=new HashSet<FAState>();
		toProcess.add(aut.getInitialState());
		result.setInitialState(result.createState());
		stMap.put(aut.getInitialState(), result.getInitialState());
		if(aut.F.contains(aut.getInitialState()))
			result.F.add(result.getInitialState());

		while(true){
			HashSet<FAState> toAdd=new HashSet<FAState>();
			Iterator<FAState> toProcess_it=toProcess.iterator();
			while(toProcess_it.hasNext()){
				FAState begin=toProcess_it.next();
				Iterator<String> sym_it=begin.nextIt();
				while(sym_it.hasNext()){
					String sym=sym_it.next();
					Iterator<FAState> end_it=begin.getNext(sym).iterator();
					while(end_it.hasNext()){
						FAState end=end_it.next();
						if(finalReachable.contains(end)){
							if(!stMap.containsKey(end)){
								toAdd.add(end);
								stMap.put(end, result.createState());
								if(aut.F.contains(end))
									result.F.add(stMap.get(end));
							}
							result.addTransition(stMap.get(begin), stMap.get(end), sym);
						}
					}
				}
			}
			if(toAdd.size()==0)
				break;
			toProcess.clear();
			toProcess.addAll(toAdd);
		}
		return result;
	}

	public int mggen=0;
	public int processed=0;
	private boolean included(){
		ArrayList<NewMetagraph> Q1=this.buildSingleCharacterNewMetagraphs();
		debug("Init:"+Q1,100);

		Iterator<NewMetagraph> Q1_it=Q1.iterator();

		while(Q1_it.hasNext()){
			NewMetagraph g=Q1_it.next();
			if(!double_graph_test(
					new MetagraphBV(
					getL(g.getLeft(),system.getInitialState().id, false), 
					getL(g.getRight(),spec.getInitialState().id,true)),
					new MetagraphBV(
					getR(g.getLeft(), true), getR(g.getRight(),false))
					)){
				return false;
			}
		}

		
		LinkedList<NewMetagraph> Next=new LinkedList<NewMetagraph>();
		TreeSet<NewMetagraph> Processed=new TreeSet<NewMetagraph>();
		TreeSet<MetagraphBV> ProcessedL= new TreeSet<MetagraphBV>();
		TreeSet<MetagraphBV> ProcessedR= new TreeSet<MetagraphBV>();
		if(Options.verbose)
			System.out.println("Next  |  Processed  |  ProcessedL ProcessedR");

		Next.addAll(Q1);
		mggen+=Q1.size();
		if(Options.verbose)
			System.out.println(Next.size()+"  |  "+Processed.size()+"  |  "+ProcessedL.size()+"  "+ProcessedR.size());
	
		while(!Next.isEmpty()&&(!Options.C1||cntL!=0)){
			if(stop)
				break;
			debug("",100);
			debug("Begin while loop",100);
			debug("Processed:"+Processed,100);
			debug("Next:"+Next,100);
			Iterator<NewMetagraph> Processed_it;
			NewMetagraph g;
			if(Options.DFS)
			    {
				g=Next.getLast();
			    }
			else if (Options.SFS)   // Smallest graph first
			    {
				g=Next.getFirst();
			        int findsmall=0;
				int minsize=g.getRight().size(); 
				while (findsmall < Next.size())
				    {
					if (Next.get(findsmall).getRight().size() < minsize)
					    {
						g=Next.get(findsmall);
						minsize=g.getRight().size();
					    }
					findsmall++;
				    }
			    }
			else
			    {
				g=Next.getFirst();  // Default BFS	
			    };
			
			debug("Pick ["+g.rpstring+"] "+g+"from Next",100);
			MetagraphBV gl=new MetagraphBV(
					getL(g.getLeft(),system.getInitialState().id,false), 
					getL(g.getRight(),spec.getInitialState().id,true));
			MetagraphBV gr=new MetagraphBV(
					getR(g.getLeft(),true), getR(g.getRight(),false));
			debug("Add "+gl+"to ProcessedL",100);
			debug("Add "+gr+"to ProcessedR",100);
			
			
			Iterator<MetagraphBV> ProcessedL_it=ProcessedL.iterator();
			while(ProcessedL_it.hasNext()){
				MetagraphBV hl=ProcessedL_it.next();
				if(!double_graph_test(hl, gr)){
					processed=Processed.size();
					return false;
				}
			}
			Iterator<MetagraphBV> ProcessedR_it=ProcessedR.iterator();
			while(ProcessedR_it.hasNext()){
				MetagraphBV hr=ProcessedR_it.next();
				if(!double_graph_test(gl, hr)){
					processed=Processed.size();
					return false;
				}
			}
			
			Next.remove(g);
			Processed.add(g);
			ProcessedL.add(gl);
			ProcessedR.add(gr);
			if(Options.verbose)
				System.out.println(Next.size()+"  |  "+Processed.size()+"  |  "+ProcessedL.size()+"  "+ProcessedR.size());

			Graph ori_L=new Graph(Options.layered);
			if(Options.C1){
				if(!Options.layered){
					Iterator<Arc> g_it=g.getLeft().iterator();
					while(g_it.hasNext()){
						Arc left=g_it.next();
						if(left.L){
							ori_L.add(left);
							left.L=false;
							cntL--;
						}
					}
				}else{
					Iterator<Integer> leftSt_it=g.getLeft().leftStateIt();
					while(leftSt_it.hasNext()){
						Iterator<Arc> g_it=g.getLeft().iterator(leftSt_it.next());
						while(g_it.hasNext()){
							Arc left=g_it.next();
							if(left.L){
								ori_L.add(left);
								left.L=false;
								cntL--;
							}
						}
					}
				}
				debug("Move"+g+"from Next to Processed with L label removed",100);
			}else{
				debug("Move"+g+"from Next to Processed",100);
			}
			
			Q1_it=Q1.iterator();
			while(Q1_it.hasNext()){
				NewMetagraph h=Q1_it.next();
				NewMetagraph f=compose(g, h,ori_L);
				if(f==null)
					continue;
				f=min_plus(f);
				debug("Produce f:["+f.rpstring+"]=["+g.rpstring +"];["+h.rpstring +"]), "+f,100);

				ArrayList<NewMetagraph> toRemove=new ArrayList<NewMetagraph>();
				ArrayList<NewMetagraph> toAdd=new ArrayList<NewMetagraph>();
				Iterator<NewMetagraph> Next_it=Next.iterator();
				while(Next_it.hasNext()){
					NewMetagraph p=Next_it.next();
					if(Options.layered)
						cleanLayered(f, p, (short) 1);
					else
						clean(f, p, (short) 1);
					if(f.getLeft().size()==0)
						break;
					
					
					Graph left=new Graph(Options.layered);
					left.addAll(p.getLeft());
					Graph right=new Graph(Options.layered);
					right.addAll(p.getRight());
					NewMetagraph p_new=new NewMetagraph(left, right,p.rpstring); 
					if(Options.layered)
						cleanLayered(p_new, f, (short) 3);
					else
						clean(p_new, f, (short) 3);
					if(p_new.compareTo(p)!=0){
						toRemove.add(p);
						if(p_new.getLeft().size()>0)
							toAdd.add(p_new);
					}
				}
				if(toAdd.size()>0){
					Next.addAll(toAdd);
					if(Options.verbose)
						System.out.println(Next.size()+"  |  "+Processed.size()+"  |  "+ProcessedL.size()+"  "+ProcessedR.size());
					toAdd.clear();
				}
				if(f.getLeft().size()==0)
					continue;
				Processed_it=Processed.iterator();
				while(Processed_it.hasNext()){
					NewMetagraph p=Processed_it.next();
					boolean regain=false;
					if(Options.layered)
						regain=!cleanLayered(f, p,(short) 2);
					else
						regain=!clean(f, p,(short) 2);
					if(regain&&Options.C1){//p regain L somewhere
						Iterator<MetagraphBV> Proc_it=ProcessedR.iterator();
						while(Proc_it.hasNext()){
							MetagraphBV right_ext=Proc_it.next();
							MetagraphBV fl=new MetagraphBV(
									getL(f.getLeft(),system.getInitialState().id,false), 
									getL(f.getRight(),spec.getInitialState().id,true));
							debug("Generate left bit vector"+fl,100);

							
							if(!double_graph_test(fl, right_ext)){
								processed=Processed.size();
								return false;
							}
						}
					}
					if(f.getLeft().size()==0)
						break;
					Graph left=new Graph(Options.layered);
					left.addAll(p.getLeft());
					Graph right=new Graph(Options.layered);
					right.addAll(p.getRight());
					NewMetagraph p_new=new NewMetagraph(left, right, p.rpstring); 
					if(Options.layered)
						cleanLayered(p_new, f, (short) 3);
					else
						clean(p_new, f, (short) 3);

					if(p_new.compareTo(p)!=0){
						toRemove.add(p);
						if(p_new.getLeft().size()>0)
							toAdd.add(p_new);
					}
				}
				if(f.getLeft().size()==0)
					continue;
				MetagraphBV left=new MetagraphBV(
						getL(f.getLeft(),system.getInitialState().id,false), 
						getL(f.getRight(),spec.getInitialState().id,true));
				MetagraphBV right=new MetagraphBV(
						getR(f.getLeft(),true), getR(f.getRight(),false));
				
				debug("Add ["+f.rpstring+"] "+f+" to Next",100);
				debug("Generate left bit vector"+left,100);
				debug("Generate right bit vector"+right,100);
				if(!double_graph_test(left,right)){
					processed=Processed.size();
					return false;
				}

				Processed.removeAll(toRemove);
				Next.removeAll(toRemove);
				Processed.addAll(toAdd);
				if(Options.verbose)
					System.out.println(Next.size()+"  |  "+Processed.size()+"  |  "+ProcessedL.size()+"  "+ProcessedR.size());

				mggen++;
//				Next_it=Next.iterator();
//				while(Next_it.hasNext()){
//					Pair<Graph, Graph> cur=Next_it.next();
//					if(cur.getRight().toString().compareToIgnoreCase(f.getRight().toString())==0){
//						merged=true;
//						cur.getLeft().addAll(f.getLeft());
//						break;
//					}
//				}
//				if(!merged)
				
				Next.add(f);
				if(Options.verbose)
					System.out.println(Next.size()+"  |  "+Processed.size()+"  |  "+ProcessedL.size()+"  "+ProcessedR.size());
			}
		}
		processed=Processed.size();
		return true;
	}			

	
	
		
	
//	private void addLabel(String type, Arc left,
//			Graph right) {
//		TreeMap<Graph, Graph> Label=null;
//		if(type.compareTo("L")==0)
//			Label =L;
//		else
//			Label =R;
//		debug("("+left+", "+right+") adds "+type+" label");
//		
//		if(Label.get(right)==null){
//			Graph empty=new Graph();
//			Label.put(right, empty);
//		}
//		Label.get(right).add(left);
//	}
//	private void delLabel(TreeMap<Graph, Graph> Label, Arc left,
//			Graph right) {
//		if(Label.get(right)==null){
//			return;
//		}
//		Label.get(right).remove(left);
//	}

	public void stopIt(){
		stop=true;
	}
}
